import json
from typing import Iterator

from dissect.sql import sqlite3
from dissect.util.ts import wintimestamp

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

HitmanAlertRecord = TargetRecordDescriptor(
    "application/av/sophos/hitman/log",
    [
        ("datetime", "ts"),
        ("string", "alert"),
        ("string", "description"),
        ("string", "details"),
    ],
)

SophosLogRecord = TargetRecordDescriptor(
    "application/av/sophos/home/log",
    [
        ("datetime", "ts"),
        ("string", "description"),
        ("path", "path"),
    ],
)


class SophosPlugin(Plugin):
    """Sophos antivirus plugin."""

    __namespace__ = "sophos"

    LOG_SOPHOS_HOME = "sysvol/ProgramData/Sophos/Clean/Logs/Clean.log"
    LOG_SOPHOS_HITMAN = "sysvol/ProgramData/HitmanPro.Alert/excalibur.db"
    MARKER_INFECTION = '{"command":"clean-threat'

    LOGS = [LOG_SOPHOS_HOME, LOG_SOPHOS_HITMAN]

    def __init__(self, target: Target) -> None:
        super().__init__(target)
        self.codepage = self.target.codepage or "ascii"

    def check_compatible(self) -> None:
        is_compatible = False
        for marker in self.LOGS:
            if self.target.fs.path(marker).exists():
                is_compatible = True
                break
        if not is_compatible:
            raise UnsupportedPluginError("No Sophos/Hitman logs found")

    @export(record=HitmanAlertRecord)
    def hitmanlogs(self) -> Iterator[HitmanAlertRecord]:
        """Return alert log records from Sophos Hitman Pro/Alert.

        Yields HitmanAlertRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp.
            alert (string): Type of Alert.
            description (string): Short description of the alert.
            details (string): Detailed description of the alert.

        Note that because Hitman also catches suspicious behaviour of
        systems, the details field might contain a lot of text, it might
        contain stracktraces etc.
        """
        try:
            fh = self.target.fs.path(self.LOG_SOPHOS_HITMAN).open("rb")
            db = sqlite3.SQLite3(fh)
            alerts = list(filter(lambda t: t.name == "Alerts", db.tables()))[0]
            for alert in alerts.rows():
                yield HitmanAlertRecord(
                    ts=wintimestamp(alert.Timestamp),  # already utc
                    alert=alert.AlertType,
                    description=alert.Description,
                    details=alert.Details,
                    _target=self.target,
                )
        except Exception as error:
            self.target.log.error(f"Error occurred during reading alerts: {error}.")

    @export(record=SophosLogRecord)
    def sophoshomelogs(self) -> Iterator[SophosLogRecord]:
        """Return log history records from Sophos Home.

        Yields SophosLogRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp.
            description (string): Short description of the alert.
            path (path): Path to the infected file (if available).

        """
        log = self.target.fs.path(self.LOG_SOPHOS_HOME).open("rt", 0, "utf-16le")
        while line := log.readline():
            if line.find(self.MARKER_INFECTION) > -1:
                try:
                    ts, json_data = line.split(" ", maxsplit=2)
                    details = json.loads(json_data)

                    path_to_infected_file = None
                    if targets := details.get("targets", None):
                        path_to_infected_file = targets[0].get("file_path", None)

                    yield SophosLogRecord(
                        ts=ts,
                        description=details.get("threat_name", details),
                        path=self.target.fs.path(path_to_infected_file),
                        _target=self.target,
                    )
                except Exception as error:
                    self.target.log.warning(f"Error: {error} on log line: {line}.")



import bz2
import gzip
import io
import os
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Iterator
from unittest.mock import Mock, patch

import pytest

from dissect.target.exceptions import (
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
    SymlinkRecursionError,
)
from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.filesystems.ntfs import NtfsFilesystemEntry
from dissect.target.helpers import fsutil
from dissect.target.target import Target


@pytest.mark.parametrize(
    ("path, alt_separator, result"),
    [
        ("/some/dir/some/file", "", "/some/dir/some/file"),
        ("/some/dir/some/file", "\\", "/some/dir/some/file"),
        ("\\some\\dir\\some\\file", "\\", "/some/dir/some/file"),
        ("/some///long\\\\dir/so\\//me\\file", "", "/some/long\\\\dir/so\\/me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "\\", "/some/long/dir/so/me/file"),
    ],
)
def test_normalize(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.normalize(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("args, alt_separator, result"),
    [
        (("/some/dir", "some/file"), "", "/some/dir/some/file"),
        (("/some/dir", "some/file"), "\\", "/some/dir/some/file"),
        (("\\some\\dir", "some\\file"), "\\", "/some/dir/some/file"),
        (("/some///long\\\\dir", "so\\//me\\file"), "", "/some/long\\\\dir/so\\/me\\file"),
        (("/some///long\\\\dir", "so\\//me\\file"), "\\", "/some/long/dir/so/me/file"),
    ],
)
def test_join(args: str, alt_separator: str, result: str) -> None:
    assert fsutil.join(*args, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path, alt_separator, result"),
    [
        ("/some/dir/some/file", "", "/some/dir/some"),
        ("/some/dir/some/file", "\\", "/some/dir/some"),
        ("\\some\\dir\\some\\file", "\\", "/some/dir/some"),
        ("/some///long\\\\dir/so\\//me\\file", "", "/some/long\\\\dir/so\\"),
        ("/some///long\\\\dir/so\\//me\\file", "\\", "/some/long/dir/so/me"),
    ],
)
def test_dirname(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.dirname(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path, alt_separator, result"),
    [
        ("/some/dir/some/file", "", "file"),
        ("/some/dir/some/file", "\\", "file"),
        ("\\some\\dir\\some\\file", "\\", "file"),
        ("/some///long\\\\dir/so\\//me\\file", "", "me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "\\", "file"),
    ],
)
def test_basename(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.basename(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path, alt_separator, result"),
    [
        ("/some/dir/some/file", "", ("/some/dir/some", "file")),
        ("/some/dir/some/file", "\\", ("/some/dir/some", "file")),
        ("\\some\\dir\\some\\file", "\\", ("/some/dir/some", "file")),
        ("/some/dir/some/", "", ("/some/dir/some", "")),
        ("/some/dir/some\\", "", ("/some/dir", "some\\")),
        ("/some/dir/some/", "\\", ("/some/dir/some", "")),
        ("\\some\\dir\\some\\", "\\", ("/some/dir/some", "")),
        ("/some///long\\\\dir/so\\//me\\file", "", ("/some/long\\\\dir/so\\", "me\\file")),
        ("/some///long\\\\dir/so\\//me\\file", "\\", ("/some/long/dir/so/me", "file")),
    ],
)
def test_split(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.split(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path, alt_separator, result"),
    [
        ("/some/dir", "", True),
        ("some/dir", "", False),
        ("\\some/dir", "", False),
        ("/some/dir", "\\", True),
        ("some/dir", "\\", False),
        ("\\some/dir", "\\", True),
    ],
)
def test_isabs(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.isabs(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path, alt_separator, result"),
    [
        ("/some/dir/../some/file", "", "/some/some/file"),
        ("/some/dir/../some/file", "\\", "/some/some/file"),
        ("\\some\\dir\\..\\some\\file", "\\", "/some/some/file"),
        ("/some///long\\..\\dir/so\\.//me\\file", "", "/some/long\\..\\dir/so\\./me\\file"),
        ("/some///long\\..\\dir/so\\.//me\\file", "\\", "/some/dir/so/me/file"),
    ],
)
def test_normpath(path: str, alt_separator: str, result: str) -> None:
    assert fsutil.normpath(path, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path, cwd, alt_separator, result"),
    [
        ("/some/dir", "", "", "/some/dir"),
        ("some/dir", "", "", "/some/dir"),
        ("\\some/dir", "", "", "/\\some/dir"),
        ("/some/dir", "", "\\", "/some/dir"),
        ("some/dir", "", "\\", "/some/dir"),
        ("\\some\\dir", "", "\\", "/some/dir"),
        ("some\\dir", "", "\\", "/some/dir"),
        ("/some/dir", "/my/cwd/", "", "/some/dir"),
        ("\\some\\dir", "\\my\\cwd\\", "\\", "/some/dir"),
        ("some/dir", "/my/cwd/", "", "/my/cwd/some/dir"),
        ("some\\dir", "/my/cwd/", "\\", "/my/cwd/some/dir"),
        ("some/dir", "/my\\cwd/", "", "/my\\cwd/some/dir"),
        ("some\\dir", "/my\\cwd/", "\\", "/my/cwd/some/dir"),
    ],
)
def test_abspath(path: str, cwd: str, alt_separator: str, result: str) -> None:
    assert fsutil.abspath(path, cwd=cwd, alt_separator=alt_separator) == result


@pytest.mark.parametrize(
    ("path, start, alt_separator, result"),
    [
        ("/some/dir/some/file", "/some/dir", "", "some/file"),
        ("/some/dir/some/file", "/some/dir", "\\", "some/file"),
        ("\\some\\dir\\some\\file", "\\some\\dir", "\\", "some/file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long/dir", "", "../../long\\\\dir/so\\/me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long\\\\dir", "", "so\\/me\\file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long/dir", "\\", "so/me/file"),
        ("/some///long\\\\dir/so\\//me\\file", "/some/long\\\\dir", "\\", "so/me/file"),
    ],
)
def test_relpath(path: str, start: str, alt_separator: str, result: str) -> None:
    assert fsutil.relpath(path, start, alt_separator=alt_separator) == result


def test_generate_addr() -> None:
    slash_path = "/some/dir/some/file"
    slash_vfs = VirtualFilesystem(alt_separator="")
    slash_target_path = fsutil.TargetPath(slash_vfs, slash_path)

    backslash_path = "\\some\\dir\\some\\file"
    backslash_vfs = VirtualFilesystem(alt_separator="\\")
    backslash_target_path = fsutil.TargetPath(backslash_vfs, backslash_path)

    assert (
        fsutil.generate_addr(slash_path, "")
        == fsutil.generate_addr(slash_path, "\\")
        == fsutil.generate_addr(backslash_path, "\\")
        == fsutil.generate_addr(slash_target_path, "")
        == fsutil.generate_addr(backslash_target_path, "")
        == fsutil.generate_addr(slash_target_path, "\\")
        == fsutil.generate_addr(backslash_target_path, "\\")
    )

    assert fsutil.generate_addr(slash_path, "") != fsutil.generate_addr(backslash_path, "")


def test_stat_result() -> None:
    with pytest.raises(TypeError):
        fsutil.stat_result([0])

    with pytest.raises(TypeError):
        fsutil.stat_result(list(range(100)))

    # ["st_mode", "st_ino", "st_dev", "st_nlink", "st_uid", "st_gid", "st_size", "_st_atime", "_st_mtime", "_st_ctime"]
    values = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    st = fsutil.stat_result(values)
    assert st[0] == st.st_mode == 0
    assert st[9] == 9
    assert st.st_ctime == 9.0
    assert st.st_ctime_ns == 9000000000
    assert st.st_blksize is None
    assert list(st) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    # [..., "st_atime", "st_mtime", "st_ctime", "st_atime_ns", "st_mtime_ns", "st_ctime_ns"]
    values += [10, 11, 12, 13, 14, 15]

    st = fsutil.stat_result(values)
    assert st[0] == st.st_mode == 0
    assert st[9] == 9
    assert st.st_ctime == 12.0
    assert st.st_ctime_ns == 15
    assert st.st_blksize is None
    assert list(st) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    my_stat = os.stat(__file__)
    st = fsutil.stat_result.copy(my_stat)
    assert st == my_stat


@pytest.fixture
def path_fs() -> Iterator[VirtualFilesystem]:
    vfs = VirtualFilesystem()

    vfs.makedirs("/some/dir")
    vfs.symlink("/some/dir/file.txt", "/some/symlink.txt")
    vfs.symlink("nonexistent", "/some/dir/link.txt")
    vfs.map_file_fh("/some/file.txt", io.BytesIO(b"content"))
    vfs.map_file_fh("/some/dir/file.txt", io.BytesIO(b""))

    yield vfs


def test_target_path_drive(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").drive == ""


def test_target_path_root(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").root == "/"


def test_target_path_anchor(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").anchor == "/"


def test_target_path_parent(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir/file.txt").parent == path_fs.path("/some/dir")


def test_target_path_parents(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/dir/file.txt")
    parents = list(path.parents)
    assert parents == [path_fs.path("/some/dir"), path_fs.path("/some"), path_fs.path("/")]
    assert [p.exists() for p in parents]
    assert all([p._fs == path_fs for p in parents])
    assert path.parents[1] == path.parents[-2] == path_fs.path("/some")


def test_target_path_name(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.name == "file.txt"


def test_target_path_suffix(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.suffix == ".txt"


def test_target_path_suffixes(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.tar.gz")
    assert path.suffixes == [".tar", ".gz"]


def test_target_path_stem(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.stem == "file"


def test_target_path_as_posix(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.as_posix() == "/some/file.txt"

    path_fs.alt_separator = "\\"
    path = path_fs.path("\\some\\file.txt")
    assert path.exists()
    assert path.as_posix() == "/some/file.txt"


def test_target_path_as_uri(path_fs: VirtualFilesystem) -> None:
    path = path_fs.path("/some/file.txt")
    assert path.as_uri() == "file:///some/file.txt"

    path_fs.alt_separator = "\\"
    path = path_fs.path("\\some\\file.txt")
    assert path.as_uri() == "file:///some/file.txt"


def test_target_path_is_absolute(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").is_absolute()
    assert not path_fs.path("some/file.txt").is_absolute()


def test_target_path_is_relative_to(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir/file.txt").is_relative_to("/some/dir")
    assert not path_fs.path("/some/dir/file.txt").is_relative_to("/some/other")


def test_target_path_is_reserved(path_fs: VirtualFilesystem) -> None:
    # We currently do not have any reserved names for TargetPath
    assert not path_fs.path("CON").is_reserved()
    assert not path_fs.path("foo").is_reserved()


def test_target_path_join(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some").joinpath("file.txt") == path_fs.path("/some/file.txt")
    assert path_fs.path("/some") / "file.txt" == path_fs.path("/some/file.txt")


def test_target_path_match(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").match("*.txt")
    assert not path_fs.path("/some/file.txt").match("*.csv")


def test_target_path_relative_to(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir/file.txt").relative_to("/some") == path_fs.path("dir/file.txt")


def test_target_path_with_name(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").with_name("new_file.txt") == path_fs.path("/some/new_file.txt")


def test_target_path_with_stem(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").with_stem("new_file") == path_fs.path("/some/new_file.txt")


def test_target_path_with_suffix(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").with_suffix(".csv") == path_fs.path("/some/file.csv")


def test_target_path_stat(path_fs: VirtualFilesystem) -> None:
    stat_result = path_fs.path("/some/file.txt").stat()
    assert stat_result.st_mode == 0o100000
    assert stat_result.st_dev == id(path_fs)
    assert stat_result.st_nlink == 1

    stat_result = path_fs.path("/some").stat()
    assert stat_result.st_mode == 0o40000
    assert stat_result.st_dev == id(path_fs)
    assert stat_result.st_nlink == 1

    assert path_fs.path("/some/symlink.txt").stat() == path_fs.path("/some/dir/file.txt").stat()


def test_target_path_lstat(path_fs: VirtualFilesystem) -> None:
    stat_result = path_fs.path("/some/symlink.txt").lstat()
    assert stat_result != path_fs.path("/some/dir/file.txt").lstat()
    assert stat_result.st_mode == 0o120000


def test_target_path_exists(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").exists()
    assert not path_fs.path("/some/other.txt").exists()


def test_target_path_glob(path_fs: VirtualFilesystem) -> None:
    assert list(path_fs.path("/some").glob("*.txt")) == [
        path_fs.path("/some/symlink.txt"),
        path_fs.path("/some/file.txt"),
    ]
    assert list(path_fs.path("/some").glob("*.csv")) == []


def test_target_path_rglob(path_fs: VirtualFilesystem) -> None:
    assert list(path_fs.path("/some").rglob("*.txt")) == [
        path_fs.path("/some/symlink.txt"),
        path_fs.path("/some/file.txt"),
        path_fs.path("/some/dir/link.txt"),
        path_fs.path("/some/dir/file.txt"),
    ]
    assert list(path_fs.path("/some").rglob("*.csv")) == []


def test_target_path_is_dir(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/dir").is_dir()
    assert not path_fs.path("/some/file.txt").is_dir()


def test_target_path_is_file(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").is_file()
    assert not path_fs.path("/some/dir").is_file()


def test_target_path_is_mount(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some").is_mount()

    mnt_vfs = VirtualFilesystem()
    mnt_vfs.makedirs("/foo")
    path_fs.mount("/mnt", mnt_vfs)

    assert path_fs.path("/mnt").is_mount()


def test_target_path_is_symlink(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").is_symlink()
    assert not path_fs.path("/some/file.txt").is_symlink()
    assert not path_fs.path("/some/dir").is_symlink()


def test_target_path_is_junction(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some").is_junction()

    mock_entry = Mock(spec=NtfsFilesystemEntry)
    mock_entry.dereference.return_value.is_mount_point.return_value = True

    path_fs.map_file_entry("/junction", mock_entry)
    assert path_fs.path("/junction").is_junction()


def test_target_path_is_socket(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_socket()


def test_target_path_is_fifo(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_fifo()


def test_target_path_is_block_device(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_block_device()


def test_target_path_is_char_device(path_fs: VirtualFilesystem) -> None:
    assert not path_fs.path("/some/file.txt").is_char_device()


def test_target_path_iterdir(path_fs: VirtualFilesystem) -> None:
    assert list(path_fs.path("/some").iterdir()) == [
        path_fs.path("/some/dir"),
        path_fs.path("/some/symlink.txt"),
        path_fs.path("/some/file.txt"),
    ]


def test_target_path_walk(path_fs: VirtualFilesystem) -> None:
    assert list(path_fs.path("/some").walk()) == [
        (path_fs.path("/some"), ["dir"], ["symlink.txt", "file.txt"]),
        (path_fs.path("/some/dir"), [], ["link.txt", "file.txt"]),
    ]


def test_target_path_open(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").open("rb").read() == b"content"
    assert path_fs.path("/some/file.txt").open("r").read() == "content"


def test_target_path_read_bytes(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").read_bytes() == b"content"


def test_target_path_read_text(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/file.txt").read_text() == "content"


def test_target_path_readlink(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").readlink() == path_fs.path("/some/dir/file.txt")


def test_target_path_resolve(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").resolve() == path_fs.path("/some/dir/file.txt")
    assert path_fs.path("/some/symlink.txt").resolve(strict=True) == path_fs.path("/some/dir/file.txt")
    assert path_fs.path("/some/foo").resolve() == path_fs.path("/some/foo")

    with pytest.raises(FileNotFoundError):
        assert path_fs.path("/some/foo").resolve(strict=True)

    with pytest.raises(FileNotFoundError):
        path_fs.path("/some/dir/link.txt").resolve(strict=True)

    with pytest.raises(NotADirectoryError):
        path_fs.path("/some/file.txt/other").resolve(strict=True)


def test_target_path_samefile(path_fs: VirtualFilesystem) -> None:
    assert path_fs.path("/some/symlink.txt").samefile(path_fs.path("/some/dir/file.txt"))
    assert not path_fs.path("/some/symlink.txt").samefile(path_fs.path("/some/file.txt"))


def test_target_path_errors(path_fs: VirtualFilesystem) -> None:
    # TargetPath sometimes emulates OSErrors to play nicely with pathlib, but other times
    # we raise our own FilesystemError variants. Ensure that all user-facing errors are our own.
    path_fs.symlink("symlink1", "symlink2")
    path_fs.symlink("symlink2", "symlink1")

    with pytest.raises(SymlinkRecursionError) as e:
        path_fs.path("symlink1/symlink2/symlink1").resolve()

    # This should raise from the final stat() call
    if sys.version_info >= (3, 10):
        assert [tb.name for tb in e.traceback[1:3]] == [
            "resolve",
            "stat",
        ]
    else:
        # In 3.9 there's no difference between these two
        assert [tb.name for tb in e.traceback[1:3]] == [
            "resolve",
            "resolve",
        ]

    with pytest.raises(SymlinkRecursionError) as e:
        path_fs.path("symlink1/symlink2/symlink1").resolve(strict=True)

    # This should raise from the inner realpath() call
    if sys.version_info >= (3, 10):
        assert [tb.frame.code.name for tb in e.traceback[1:3]] == [
            "resolve",
            "realpath",
        ]
    else:
        # In 3.9 there's no difference between these two
        assert [tb.frame.code.name for tb in e.traceback[1:3]] == [
            "resolve",
            "resolve",
        ]

    with pytest.raises(NotASymlinkError):
        path_fs.path("some/file.txt").readlink()

    with pytest.raises(NotADirectoryError):
        path_fs.path("some/file.txt/dir").stat()


def test_target_path_not_implemented(path_fs: VirtualFilesystem) -> None:
    # TargetPath can't do some things, such as write actions or stuff related to a "current" user or path
    # Ensure all those methods properly error
    with pytest.raises(NotImplementedError):
        assert path_fs.path().cwd()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().home()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().expanduser()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().absolute()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().owner()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().group()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().chmod(0o777)

    with pytest.raises(NotImplementedError):
        assert path_fs.path().lchmod(0o777)

    with pytest.raises(NotImplementedError):
        assert path_fs.path().rename("foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().replace("foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().symlink_to("foo")

    if sys.version_info >= (3, 10):
        with pytest.raises(NotImplementedError):
            assert path_fs.path().hardlink_to("foo")
    else:
        with pytest.raises(NotImplementedError):
            assert path_fs.path().link_to("foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().mkdir()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().rmdir()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().touch()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().unlink()

    with pytest.raises(NotImplementedError):
        assert path_fs.path().write_bytes(b"foo")

    with pytest.raises(NotImplementedError):
        assert path_fs.path().write_text("foo")


@pytest.mark.parametrize(
    ("path, alt_separator, result"),
    [
        ("/some/dir/some/file", "", ("/", "some", "dir", "some", "file")),
        ("/some/dir//some/file", "", ("/", "some", "dir", "some", "file")),
        ("/some/dir\\some/file", "", ("/", "some", "dir\\some", "file")),
        ("\\some\\dir\\some\\file", "\\", ("/", "some", "dir", "some", "file")),
        ("/some/dir/some/file", "\\", ("/", "some", "dir", "some", "file")),
        ("\\some\\dir\\\\some\\file", "\\", ("/", "some", "dir", "some", "file")),
        ("\\some\\dir/some\\file", "\\", ("/", "some", "dir", "some", "file")),
    ],
)
def test_pure_dissect_path__from_parts(path: str, alt_separator: str, result: tuple[str]) -> None:
    vfs = VirtualFilesystem(alt_separator=alt_separator)
    pure_dissect_path = fsutil.PureDissectPath(vfs, path)

    assert pure_dissect_path.parts == result


@pytest.mark.parametrize(
    ("alt_separator"),
    ["/", "\\"],
)
@pytest.mark.parametrize(
    ("case_sensitive"),
    [True, False],
)
def test_pure_dissect_path__from_parts_flavour(alt_separator: str, case_sensitive: bool) -> None:
    vfs = VirtualFilesystem(alt_separator=alt_separator, case_sensitive=case_sensitive)
    pure_dissect_path = fsutil.PureDissectPath(vfs, "/some/dir")

    assert pure_dissect_path._flavour.altsep == alt_separator
    assert pure_dissect_path._flavour.case_sensitive == case_sensitive


def test_pure_dissect_path__from_parts_no_fs_exception() -> None:
    with pytest.raises(TypeError):
        fsutil.PureDissectPath(Mock(), "/some/dir")


@pytest.mark.parametrize(
    ("file_name, compressor, content"),
    [
        ("plain", lambda x: x, b"plain\ncontent"),
        ("comp.gz", gzip.compress, b"gzip\ncontent"),
        ("comp_gz", gzip.compress, b"gzip\ncontent"),
        ("comp.bz2", bz2.compress, b"bz2\ncontent"),
        ("comp_bz2", bz2.compress, b"bz2\ncontent"),
    ],
)
def test_open_decompress(file_name: str, compressor: Callable, content: bytes) -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_fh(file_name, io.BytesIO(compressor(content)))
    assert fsutil.open_decompress(vfs.path(file_name)).read() == content


def test_open_decompress_text_modes() -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_fh("test", io.BytesIO(b"zomgbbq"))

    fh = fsutil.open_decompress(vfs.path("test"))
    assert not isinstance(fh, io.TextIOWrapper)

    fh = fsutil.open_decompress(vfs.path("test"), "r")
    assert isinstance(fh, io.TextIOWrapper)
    assert fh.encoding == "UTF-8"
    assert fh.errors == "backslashreplace"

    fh = fsutil.open_decompress(vfs.path("test"), "r", errors=None)
    assert isinstance(fh, io.TextIOWrapper)
    assert fh.encoding == "UTF-8"
    assert fh.errors == "strict"

    fh = fsutil.open_decompress(vfs.path("test"), "r", encoding="ascii")
    assert isinstance(fh, io.TextIOWrapper)
    assert fh.encoding == "ascii"
    assert fh.errors == "backslashreplace"


def test_reverse_readlines() -> None:
    vfs = VirtualFilesystem()

    expected_range_reverse = ["99"] + [f"{i}\n" for i in range(98, -1, -1)]

    vfs.map_file_fh("file_n", io.BytesIO("\n".join(map(str, range(0, 100))).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_n").open("rt"))) == expected_range_reverse

    vfs.map_file_fh("file_r", io.BytesIO("\r".join(map(str, range(0, 100))).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_r").open("rt"))) == expected_range_reverse

    vfs.map_file_fh("file_rn", io.BytesIO("\r\n".join(map(str, range(0, 100))).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_rn").open("rt"))) == expected_range_reverse

    vfs.map_file_fh("file_multi", io.BytesIO("🦊\n🦊🦊\n🦊🦊🦊".encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_multi").open("rt"))) == ["🦊🦊🦊", "🦊🦊\n", "🦊\n"]

    vfs.map_file_fh("file_multi_long", io.BytesIO((("🦊" * 8000) + ("a" * 200) + "\n🦊🦊\n🦊🦊🦊").encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_multi_long").open("rt"))) == [
        "🦊🦊🦊",
        "🦊🦊\n",
        ("🦊" * 8000) + ("a" * 200) + "\n",
    ]

    vfs.map_file_fh("file_multi_long_single", io.BytesIO((("🦊" * 8000) + ("a" * 200)).encode()))
    assert list(fsutil.reverse_readlines(vfs.path("file_multi_long_single").open("rt"))) == [("🦊" * 8000) + ("a" * 200)]

    vfs.map_file_fh("empty", io.BytesIO(b""))
    assert list(fsutil.reverse_readlines(vfs.path("empty").open("rt"))) == []

    broken_content = (b"foobar\r\n" * 2) + (b"\xc2broken\r\n") + (b"barfoo\r\n" * 2)
    vfs.map_file_fh("file_multi_broken", io.BytesIO(broken_content))
    with pytest.raises(
        UnicodeDecodeError, match="'UTF-8' codec can't decode bytes in position 0-17: failed to decode line"
    ):
        assert list(fsutil.reverse_readlines(vfs.path("file_multi_broken").open("rt"))) == ["barfoo\n", "barfoo\n"]

    assert list(fsutil.reverse_readlines(vfs.path("file_multi_broken").open("rt", errors="backslashreplace"))) == [
        "barfoo\n",
        "barfoo\n",
        "\\xc2broken\n",
        "foobar\n",
        "foobar\n",
    ]


@pytest.fixture
def xattrs() -> dict[str, bytes]:
    return {"some_key": b"some_value"}


@pytest.fixture
def listxattr_spec(xattrs: dict[str, str]) -> dict[str, Any]:
    # listxattr() is only available on Linux
    attr_names = list(xattrs.keys())

    if hasattr(os, "listxattr"):
        spec = {
            "create": False,
            "autospec": True,
            "return_value": attr_names,
        }
    else:
        spec = {
            "create": True,
            "return_value": attr_names,
        }

    return spec


@pytest.fixture
def getxattr_spec(xattrs: dict[str, str]) -> dict[str, Any]:
    # getxattr() is only available on Linux
    attr_name = list(xattrs.keys())[0]
    attr_value = xattrs.get(attr_name)

    if hasattr(os, "getxattr"):
        spec = {
            "create": False,
            "autospec": True,
            "return_value": attr_value,
        }
    else:
        spec = {
            "create": True,
            "return_value": attr_value,
        }

    return spec


@pytest.mark.parametrize(
    "follow_symlinks",
    [
        True,
        False,
    ],
)
def test_fs_attrs(
    xattrs: dict[str, bytes], listxattr_spec: dict[str, Any], getxattr_spec: dict[str, Any], follow_symlinks: bool
) -> None:
    with patch("os.listxattr", **listxattr_spec) as listxattr:
        with patch("os.getxattr", **getxattr_spec) as getxattr:
            path = "/some/path"
            attr_name = list(xattrs.keys())[0]

            assert fsutil.fs_attrs(path, follow_symlinks=follow_symlinks) == xattrs
            listxattr.assert_called_with(path, follow_symlinks=follow_symlinks)
            getxattr.assert_called_with(path, attr_name, follow_symlinks=follow_symlinks)


@contextmanager
def no_listxattr() -> Iterator[None]:
    if not hasattr(os, "listxattr"):
        yield
        return
    listxattr = os.listxattr
    try:
        del os.listxattr
        yield
    finally:
        os.listxattr = listxattr


def test_fs_attrs_no_os_listxattr() -> None:
    with no_listxattr():
        assert fsutil.fs_attrs("/some/path") == {}


def test_target_path_checks_dirfs(tmp_path: Path, target_win: Target) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b"dummy")
        tf.close()
        tmpfile_name = os.path.basename(tf.name)

        fs = DirectoryFilesystem(path=tmp_path)
        target_win.filesystems.add(fs)
        target_win.fs.mount("Z:\\", fs)
        assert target_win.fs.path(f"Z:\\{tmpfile_name}").is_file()
        assert not target_win.fs.path(f"Z:\\{tmpfile_name}\\some").exists()
        assert not target_win.fs.path(f"Z:\\{tmpfile_name}\\some").is_file()


def test_target_path_checks_mapped_dir(tmp_path: Path, target_win: Target) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b"dummy")
        tf.close()
        tmpfile_name = os.path.basename(tf.name)

        target_win.filesystems.entries[0].map_dir("test-dir", tmp_path)
        assert target_win.fs.path("C:\\test-dir\\").is_dir()
        assert not target_win.fs.path("C:\\test-dir\\").is_file()

        assert target_win.fs.path(f"C:\\test-dir\\{tmpfile_name}").is_file()
        assert not target_win.fs.path(f"C:\\test-dir\\{tmpfile_name}\\some").is_file()


def test_target_path_checks_virtual() -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_entry("file", VirtualFile(vfs, "file", None))
    assert not vfs.path("file/test").exists()


def test_target_path_backslash_normalisation(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b"dummy")
        tf.close()

        fs_win.map_dir("windows/system32/", tmp_path)
        fs_win.map_file("windows/system32/somefile.txt", tf.name)

        results = list(target_win.fs.path("/").glob("C:\\windows\\system32\\some*.txt"))
        assert len(results) == 1

        results = list(target_win.fs.path("/").glob("sysvol/windows/system32/some*.txt"))
        assert len(results) == 1


@pytest.fixture
def glob_fs() -> VirtualFilesystem:
    vfs = VirtualFilesystem()
    paths = [
        "foo/bar/bla",
        "moo/bar/bla",
        "bar/bla",
    ]
    files = [
        "file.txt",
        "file.ini",
        "other.txt",
    ]
    special_files = [
        "lololo",
        "system.dat",
        "data.tgz",
    ]

    for idx, path in enumerate(paths):
        vfs.makedirs(path)
        for file in files:
            vfs.map_file_entry(f"/{path}/{file}", VirtualFile(vfs, f"{path}/{file}", None))
        special_file = special_files[idx]
        vfs.map_file_entry(f"/{path}/{special_file}", VirtualFile(vfs, f"{path}/{special_file}", None))

    return vfs


@pytest.mark.parametrize(
    ("start_path, pattern, results"),
    [
        ("/", "foo/bar/bla/file.*", ["foo/bar/bla/file.ini", "foo/bar/bla/file.txt"]),
        ("/", "foo/bar/*/file.ini", ["foo/bar/bla/file.ini"]),
        ("/", "foo/bar/*/file.*", ["foo/bar/bla/file.ini", "foo/bar/bla/file.txt"]),
        ("/", "*/bar/bla/file.ini", ["foo/bar/bla/file.ini", "moo/bar/bla/file.ini"]),
        ("/", "*/bar/bla/*.ini", ["foo/bar/bla/file.ini", "moo/bar/bla/file.ini"]),
        ("/foo", "*/bla/file.ini", ["foo/bar/bla/file.ini"]),
        ("/foo", "*/bla/*.ini", ["foo/bar/bla/file.ini"]),
        ("/", "boo/bla/*", []),
    ],
)
def test_glob_ext(glob_fs: VirtualFilesystem, start_path: str, pattern: str, results: list[str]) -> None:
    start_entry = glob_fs.get(start_path)
    entries = fsutil.glob_ext(start_entry, pattern)

    entries = sorted([entry.path for entry in entries])
    assert entries == sorted(results)



from dissect.hypervisor.descriptor.vbox import VBox

from dissect.target import container
from dissect.target.loader import Loader


class VBoxLoader(Loader):
    """Load Oracle VirtualBox files."""

    def __init__(self, path, **kwargs):
        path = path.resolve()

        super().__init__(path)
        with path.open("r") as fh:
            self.vbox = VBox(fh)
        self.base_dir = path.parent

    @staticmethod
    def detect(path):
        return path.suffix.lower() == ".vbox"

    def map(self, target):
        for disk in self.vbox.disks():
            parent = self.open(disk)
            target.disks.add(container.open(parent))

    def open(self, path):
        return self.base_dir.joinpath(path).open("rb")



import posixpath
from datetime import datetime
from ipaddress import ip_address, ip_network
from typing import Counter
from unittest.mock import MagicMock, patch

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.general.network import UnixInterfaceRecord
from dissect.target.plugins.os.unix.linux.network import (
    LinuxNetworkConfigParser,
    LinuxNetworkPlugin,
    NetworkManagerConfigParser,
    SystemdNetworkConfigParser,
)


def test_networkmanager_parser(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    fixture_dir = "tests/_data/plugins/os/unix/linux/NetworkManager/"
    fs_linux.map_dir("/etc/NetworkManager/system-connections", fixture_dir)

    network_manager_config_parser = NetworkManagerConfigParser(target_linux)
    interfaces = list(network_manager_config_parser.interfaces())

    assert len(interfaces) == 2
    wired, wireless = interfaces

    assert wired.name == "enp0s3"
    assert wired.type == "ethernet"
    assert wired.mac == ["08:00:27:5B:4A:EB"]
    assert Counter(wired.ip) == Counter([ip_address("192.168.2.138"), ip_address("10.1.1.10")])
    assert wired.dns == [ip_address("88.88.88.88")]
    assert Counter(wired.gateway) == Counter(
        [ip_address("192.168.2.2"), ip_address("2620:52:0:2219:222:68ff:fe11:5403"), ip_address("192.168.2.3")]
    )
    assert Counter(wired.network) == Counter([ip_network("192.168.2.0/24"), ip_network("10.1.0.0/16")])
    assert not wired.dhcp_ipv4
    assert not wired.dhcp_ipv6
    assert wired.enabled is None
    assert wired.last_connected == datetime.fromisoformat("2024-10-29 07:59:54+00:00")
    assert Counter(wired.vlan) == Counter([10, 11])
    assert wired.source == "/etc/NetworkManager/system-connections/wired-static.nmconnection"
    assert wired.configurator == "NetworkManager"

    assert wireless.name == "wlp2s0"
    assert wireless.type == "wifi"
    assert wireless.mac == []
    assert wireless.ip == []
    assert wireless.dns == []
    assert wireless.gateway == []
    assert wireless.network == []
    assert wireless.dhcp_ipv4
    assert wireless.dhcp_ipv6
    assert wireless.enabled is None
    assert wireless.last_connected is None
    assert wireless.vlan == []
    assert wireless.source == "/etc/NetworkManager/system-connections/wireless.nmconnection"
    assert wireless.configurator == "NetworkManager"


def test_systemd_network_parser(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    fixture_dir = "tests/_data/plugins/os/unix/linux/systemd.network/"
    fs_linux.makedirs("/etc/systemd/network")

    fs_linux.map_file(
        "/etc/systemd/network/20-wired-static.network", posixpath.join(fixture_dir, "20-wired-static.network")
    )
    fs_linux.map_file(
        "/etc/systemd/network/30-wired-static-complex.network",
        posixpath.join(fixture_dir, "30-wired-static-complex.network"),
    )
    fs_linux.map_file(
        "/usr/lib/systemd/network/40-wireless.network", posixpath.join(fixture_dir, "40-wireless.network")
    )
    fs_linux.map_file(
        "/run/systemd/network/40-wireless-ipv4.network", posixpath.join(fixture_dir, "40-wireless-ipv4.network")
    )
    fs_linux.map_file(
        "/usr/local/lib/systemd/network/40-wireless-ipv6.network",
        posixpath.join(fixture_dir, "40-wireless-ipv6.network"),
    )
    fs_linux.map_file("/etc/systemd/network/20-vlan.netdev", posixpath.join(fixture_dir, "20-vlan.netdev"))
    fs_linux.map_file("/etc/systemd/network/20-vlan2.netdev", posixpath.join(fixture_dir, "20-vlan2.netdev"))

    systemd_network_config_parser = SystemdNetworkConfigParser(target_linux)
    interfaces = list(systemd_network_config_parser.interfaces())

    assert len(interfaces) == 5
    wired_static, wired_static_complex, wireless, wireless_ipv4, wireless_ipv6 = interfaces

    assert wired_static.name == "enp1s0"
    assert wired_static.type is None
    assert wired_static.mac == ["aa::bb::cc::dd::ee::ff"]
    assert wired_static.ip == [ip_address("10.1.10.9")]
    assert wired_static.dns == [ip_address("10.1.10.1")]
    assert wired_static.gateway == [ip_address("10.1.10.1")]
    assert wired_static.network == [ip_network("10.1.10.0/24")]
    assert not wired_static.dhcp_ipv4
    assert not wired_static.dhcp_ipv6
    assert wired_static.enabled is None
    assert wired_static.last_connected is None
    assert Counter(wired_static.vlan) == Counter([100, 101])
    assert wired_static.source == "/etc/systemd/network/20-wired-static.network"
    assert wired_static.configurator == "systemd-networkd"

    assert wired_static_complex.name == "enp1s0"
    assert wired_static_complex.type == "ether"
    assert Counter(wired_static_complex.mac) == Counter(
        ["aa::bb::cc::dd::ee::ff", "ff::ee::dd::cc::bb::aa", "cc::ff::bb::aa::dd", "bb::aa::dd::cc::ff"]
    )
    assert Counter(wired_static_complex.ip) == Counter([ip_address("10.1.10.9"), ip_address("10.1.9.10")])
    assert Counter(wired_static_complex.dns) == Counter(
        [ip_address("10.1.10.1"), ip_address("10.1.10.2"), ip_address("1111:2222::3333")]
    )
    assert Counter(wired_static_complex.gateway) == Counter(
        [ip_address("10.1.6.3"), ip_address("10.1.10.2"), ip_address("10.1.9.3")]
    )
    assert Counter(wired_static_complex.network) == Counter([ip_network("10.1.0.0/16"), ip_network("10.1.9.0/24")])
    assert not wired_static_complex.dhcp_ipv4
    assert not wired_static_complex.dhcp_ipv6
    assert wired_static_complex.enabled is None
    assert wired_static_complex.last_connected is None
    assert wired_static_complex.vlan == []
    assert wired_static_complex.source == "/etc/systemd/network/30-wired-static-complex.network"
    assert wired_static_complex.configurator == "systemd-networkd"

    assert wireless.name == "wlp2s0"
    assert wireless.type == "wifi"
    assert wireless.mac == []
    assert wireless.ip == []
    assert wireless.dns == []
    assert wireless.gateway == []
    assert wireless.network == []
    assert wireless.dhcp_ipv4
    assert wireless.dhcp_ipv6
    assert wireless.enabled is None
    assert wireless.last_connected is None
    assert wired_static_complex.vlan == []
    assert wireless.source == "/usr/lib/systemd/network/40-wireless.network"
    assert wired_static_complex.configurator == "systemd-networkd"

    assert wireless_ipv4.dhcp_ipv4
    assert not wireless_ipv4.dhcp_ipv6
    assert wireless_ipv4.source == "/run/systemd/network/40-wireless-ipv4.network"

    assert not wireless_ipv6.dhcp_ipv4
    assert wireless_ipv6.dhcp_ipv6
    assert wireless_ipv6.source == "/usr/local/lib/systemd/network/40-wireless-ipv6.network"


def test_linux_network_plugin_interfaces(target_linux: Target) -> None:
    """Assert that the LinuxNetworkPlugin aggregates from all Config Parsers."""

    MockLinuxConfigParser1: LinuxNetworkConfigParser = MagicMock()
    MockLinuxConfigParser1.return_value.interfaces.return_value = []

    MockLinuxConfigParser2: LinuxNetworkConfigParser = MagicMock()
    MockLinuxConfigParser2.return_value.interfaces.return_value = [UnixInterfaceRecord()]

    with patch(
        "dissect.target.plugins.os.unix.linux.network.MANAGERS", [MockLinuxConfigParser1, MockLinuxConfigParser2]
    ):
        linux_network_plugin = LinuxNetworkPlugin(target_linux)
        interfaces = list(linux_network_plugin.interfaces())

        assert len(interfaces) == 1
        MockLinuxConfigParser1.return_value.interfaces.assert_called_once()
        MockLinuxConfigParser2.return_value.interfaces.assert_called_once()



from io import BytesIO

import pytest

from dissect.target.plugins.os.unix.linux.fortios._keys import KERNEL_KEY_MAP
from dissect.target.plugins.os.unix.linux.fortios._os import (
    decrypt_rootfs,
    key_iv_for_kernel_hash,
)


def test_kernel_key_map() -> None:
    # Ensure that the kernel key map is valid
    for kernel_hash, key in KERNEL_KEY_MAP.items():
        # test if the kernel hash is a valid hex string
        assert bytes.fromhex(kernel_hash)

        # test if the kernel hash is a valid length (sha256)
        assert len(bytes.fromhex(kernel_hash)) == 32

        # test if the key is a valid hex string
        assert bytes.fromhex(key)

        # test if key is valid length, 32 bytes for KDF, and 48 bytes for static key + IV
        assert len(bytes.fromhex(key)) in [32, 48]


def test_key_iv_for_kernel_hash() -> None:
    # test FFW_1801F-v7.4.2.F-build2571-FORTINET (KDF)
    key, iv = key_iv_for_kernel_hash("d719f7fd533d05efb872907cf3711d0d895750d288b856ce70fefecbd7ace482")
    assert key == bytes.fromhex("39ef9ceb4262b49252164a4558b14a9b006d91a5247f5c797af281fade2198a8")
    assert iv == bytes.fromhex("f30a9e100417e2c390b763d2be2f2d03")

    # test FFW_3980E-v7.0.14.M-build0601-FORTINET (static key + IV)
    key, iv = key_iv_for_kernel_hash("a494ec1713ab75a5ab58a847f096951e2de7ba899bef1a9a88a9c94d8efc4749")
    assert key == bytes.fromhex("bb48ece8482e277f307479b8923796aed3b536e83f5fadc0d758b36192626762")
    assert iv == bytes.fromhex("5394687ae6c679a74c7901267dfb9bb3")

    # test unknown hash
    with pytest.raises(ValueError, match="No known decryption keys for kernel hash: .*"):
        key_iv_for_kernel_hash("12345")


def test_decrypt_rootfs() -> None:
    # test decrypt of rootfs.gz header (FWF_81F_2R_POE-v7.4.3.F-build2573-FORTINET)
    encrypted_rootfs_header = bytes.fromhex("3ccb 7d85 b9b0 4c8e 8c92 36d4 1d9c c48c")
    key = bytes.fromhex("b9c77cfca5c3f4fe543b5b861b5eeab61b0bfd23fa93f52f5cd428bb5567ec37")
    iv = bytes.fromhex("25c9578ca8d04f8c55009ae41657d7dd")
    fh = decrypt_rootfs(BytesIO(encrypted_rootfs_header), key, iv)
    assert fh.read(16) == b"\x1f\x8b\x08\x00J\xd6\xbbe\x00\x03\xa4\xb6S\x900\x00"

    # test bad decrypt
    with pytest.raises(ValueError, match="Failed to decrypt: No gzip magic header found."):
        decrypt_rootfs(BytesIO(encrypted_rootfs_header), key[::-1], iv[::-1])



from __future__ import annotations

import stat
from typing import BinaryIO, Iterator, Optional

from dissect.fat import exfat
from dissect.util.stream import RunlistStream
from dissect.util.ts import UTC, dostimestamp

from dissect.target.exceptions import FileNotFoundError, NotADirectoryError
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

ExfatFileTree = tuple[exfat.c_exfat.FILE, dict[str, Optional["ExfatFileTree"]]]


class ExfatFilesystem(Filesystem):
    __type__ = "exfat"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, case_sensitive=False, alt_separator="\\", *args, **kwargs)
        self.exfat = exfat.ExFAT(fh)
        self.cluster_size = self.exfat.cluster_size

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return fh.read(11)[3:] == b"EXFAT   "

    def get(self, path: str) -> ExfatFilesystemEntry:
        """Returns a ExfatFilesystemEntry object corresponding to the given pathname"""
        return ExfatFilesystemEntry(self, path, self._get_entry(path))

    def _get_entry(self, path: str, root: Optional[ExfatFileTree] = None) -> ExfatFileTree:
        dirent = root if root is not None else self.exfat.files["/"]

        # Programmatically we will often use the `/` separator, so replace it
        # with the native path separator of exFAT `/` is an illegal character
        # in exFAT filenames, so it's safe to replace
        parts = path.replace("/", "\\").split("\\")

        for part in parts:
            if not part:
                continue

            file_tree = dirent[1]
            if file_tree is None:
                raise NotADirectoryError(f"Not a directory: {path}")

            for entry_name, entry_file_tree in file_tree.items():
                if entry_name.upper() == part.upper():
                    dirent = entry_file_tree
                    break
            else:
                raise FileNotFoundError(f"File not found: {path}")

        return dirent


class ExfatFilesystemEntry(FilesystemEntry):
    def __init__(
        self,
        fs: ExfatFilesystem,
        path: str,
        entry: ExfatFileTree,
    ):
        super().__init__(fs, path, entry)
        self.size = self.entry[0].stream.data_length
        self.cluster = self.entry[0].stream.location

    def get(self, path: str) -> ExfatFilesystemEntry:
        """Get a filesystem entry relative from the current one."""
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return ExfatFilesystemEntry(self.fs, full_path, self.fs._get_entry(path, self.entry))

    def open(self) -> BinaryIO:
        """Returns file handle (file like object)"""
        if self.entry[0].stream.flags.not_fragmented:
            runlist = self.fs.exfat.runlist(self.cluster, True, self.size)
        else:
            runlist = self.fs.exfat.runlist(self.cluster, False)
        fh = RunlistStream(self.fs.exfat.filesystem, runlist, self.size, self.fs.cluster_size)
        return fh

    def _iterdir(self) -> Iterator[tuple[str, ExfatFileTree]]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry_name, entry_file_tree in self.entry[1].items():
            if entry_name in (".", ".."):
                continue
            yield (entry_name, entry_file_tree)

    def iterdir(self) -> Iterator[str]:
        """List the directory contents of a directory. Returns a generator of strings."""
        for entry_name, _ in self._iterdir():
            yield entry_name

    def scandir(self) -> Iterator[ExfatFilesystemEntry]:
        """List the directory contents of this directory. Returns a generator of filesystem entries."""
        for entry_name, entry_file_tree in self._iterdir():
            path = fsutil.join(self.path, entry_name, alt_separator=self.fs.alt_separator)
            yield ExfatFilesystemEntry(self.fs, path, entry_file_tree)

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return False

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        return bool(self.entry[0].metadata.attributes.directory)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        return not self.is_dir()

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        fe = self.entry[0]
        size = fe.stream.data_length
        addr = fe.stream.location

        # exfat stores additional offsets for ctime and mtime to get 10 millisecond precision
        fe = fe.metadata

        # all timestamps are recorded in local time. the utc offset (of the system generating the timestamp in question)
        # is recorded in the associated tz byte
        c_tz = UTC(self.fs.exfat._utc_timezone(fe.create_timezone))  # noqa
        m_tz = UTC(self.fs.exfat._utc_timezone(fe.modified_timezone))  # noqa
        a_tz = UTC(self.fs.exfat._utc_timezone(fe.access_timezone))  # noqa

        ctime = dostimestamp(fe.create_time, fe.create_offset).replace(tzinfo=c_tz)
        mtime = dostimestamp(fe.modified_time, fe.modified_offset).replace(tzinfo=m_tz)
        atime = dostimestamp(fe.access_time).replace(tzinfo=a_tz)

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = [
            (stat.S_IFDIR if self.is_dir() else stat.S_IFREG) | 0o777,
            addr,
            id(self.fs),
            0,
            0,
            0,
            size,
            atime.timetuple().timestamp(),
            mtime.timetuple().timestamp(),
            ctime.timetuple().timestamp(),
        ]
        return fsutil.stat_result(st_info)



from __future__ import annotations

import plistlib
from typing import Iterator, Optional

from flow.record.fieldtypes import posix_path

from dissect.target.filesystem import Filesystem
from dissect.target.helpers.record import UnixUserRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin
from dissect.target.target import Target


class MacPlugin(BsdPlugin):
    VERSION = "/System/Library/CoreServices/SystemVersion.plist"
    GLOBAL = "/Library/Preferences/.GlobalPreferences.plist"
    SYSTEM = "/Library/Preferences/SystemConfiguration/preferences.plist"

    @classmethod
    def detect(cls, target: Target) -> Optional[Filesystem]:
        for fs in target.filesystems:
            if fs.exists("/Library") and fs.exists("/Applications"):
                return fs

        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> MacPlugin:
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(property=True)
    def hostname(self) -> Optional[str]:
        try:
            preferences = plistlib.load(self.target.fs.path(self.SYSTEM).open())
            return preferences["System"]["System"]["ComputerName"]

        except FileNotFoundError:
            pass

    @export(property=True)
    def ips(self) -> Optional[list[str]]:
        return list(set(map(str, self.target.network.ips())))

    @export(property=True)
    def version(self) -> Optional[str]:
        try:
            systemVersion = plistlib.load(self.target.fs.path(self.VERSION).open())
            productName = systemVersion["ProductName"]
            productUserVisibleVersion = systemVersion["ProductUserVisibleVersion"]
            productBuildVersion = systemVersion["ProductBuildVersion"]
            return f"{productName} {productUserVisibleVersion} ({productBuildVersion})"
        except FileNotFoundError:
            pass

    @export(record=UnixUserRecord)
    def users(self) -> Iterator[UnixUserRecord]:
        try:
            for path in self.target.fs.path("/var/db/dslocal/nodes/Default/users/").glob("*.plist"):
                user = plistlib.load(path.open())

                # The home directory of a user account can be null,
                # but a user account can also have multiply home directories e.g. the root account.
                # https://developer.apple.com/documentation/foundation/filemanager/1642853-homedirectory/
                for home_dir in user.get("home", [None]):
                    yield UnixUserRecord(
                        name=user.get("name", [None])[0],
                        passwd=user.get("passwd", [None])[0],
                        uid=user.get("uid", [None])[0],
                        gid=user.get("gid", [None])[0],
                        gecos=user.get("realname", [None])[0],
                        home=posix_path(home_dir) if home_dir else None,
                        shell=user.get("shell", [None])[0],
                        source=path,
                    )
        except FileNotFoundError:
            pass

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.OSX.value

    @export(property=True)
    def architecture(self) -> Optional[str]:
        # OS-X uses Mach-O binary format. We should implement something similar
        # to the Unix architecture() function but for Mach-O libraries.
        pass



from typing import Union

import pytest

from dissect.target.plugins.general.default import DefaultPlugin
from dissect.target.target import Target


@pytest.mark.parametrize(
    "method_name, result",
    [
        ("hostname", None),
        ("ips", []),
        ("version", None),
        ("os", "default"),
        ("architecture", None),
    ],
)
def test_default_plugin_property_methods(
    target_default: Target,
    method_name: str,
    result: Union[None, str, list],
) -> str:
    os_plugin = DefaultPlugin(target_default)
    attr = getattr(os_plugin, method_name)
    assert attr == result



from io import BytesIO
from pathlib import Path

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.shadow import ShadowPlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_unix_shadow(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    shadow_file = absolute_path("_data/plugins/os/unix/shadow/shadow")
    fs_unix.map_file("/etc/shadow", shadow_file)
    target_unix_users.add_plugin(ShadowPlugin)

    results = list(target_unix_users.passwords())
    assert len(results) == 1
    assert results[0].name == "test"
    assert (
        results[0].crypt
        == "$6$oLHns1qc.C3DoQ8c$temOg6X.UF5Ly3gM03cGnBLib30mv8J49dUI.w9.EHTnO4R467zyKbfBnmTa5IIvDr5mRXFoJVBGKF6QuFDpo1"
    )  # noqa E501
    assert results[0].salt == "oLHns1qc.C3DoQ8c"
    assert (
        results[0].hash == "temOg6X.UF5Ly3gM03cGnBLib30mv8J49dUI.w9.EHTnO4R467zyKbfBnmTa5IIvDr5mRXFoJVBGKF6QuFDpo1"
    )  # noqa E501
    assert results[0].algorithm == "sha512"
    assert results[0].crypt_param is None
    assert results[0].last_change == "18963"
    assert results[0].min_age == 0
    assert results[0].max_age == 99999
    assert results[0].warning_period == 7
    assert results[0].inactivity_period == ""
    assert results[0].expiration_date == ""
    assert results[0].unused_field == ""


def test_unix_shadow_backup_file(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """test if both the shadow file and shadow backup file are read and returns unique hash+user combinations"""
    shadow_file = absolute_path("_data/plugins/os/unix/shadow/shadow")
    fs_unix.map_file("/etc/shadow", shadow_file)

    first_entry = Path(shadow_file).open("rb").read()
    other_entry = first_entry.replace(b"test", b"other-user")
    duplicate_entry = first_entry
    fs_unix.map_file_fh("/etc/shadow-", BytesIO(first_entry + other_entry + duplicate_entry))

    results = list(target_unix_users.passwords())
    assert len(results) == 2
    assert results[0].name == "test"
    assert results[1].name == "other-user"
    assert results[0].hash == results[1].hash



from __future__ import annotations

import logging

from dissect.target.filesystem import Filesystem
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.bsd.osx._os import MacPlugin
from dissect.target.plugins.os.unix.linux.network_managers import (
    LinuxNetworkManager,
    parse_unix_dhcp_log_messages,
)
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.target import Target

log = logging.getLogger(__name__)


class LinuxPlugin(UnixPlugin, LinuxNetworkManager):
    def __init__(self, target: Target):
        super().__init__(target)
        self.network_manager = LinuxNetworkManager(target)
        self.network_manager.discover()

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if (
                (fs.exists("/var") and fs.exists("/etc") and fs.exists("/opt"))
                or (fs.exists("/sys/module") or fs.exists("/proc/sys"))
            ) and not (MacPlugin.detect(target) or WindowsPlugin.detect(target)):
                return fs

    @export(property=True)
    def ips(self) -> list[str]:
        """Returns a list of static IP addresses and DHCP lease IP addresses found on the host system."""
        ips = set()

        for ip_set in self.network_manager.get_config_value("ips"):
            ips.update(ip_set)

        for ip in parse_unix_dhcp_log_messages(self.target, iter_all=False):
            ips.add(ip)

        return list(ips)

    @export(property=True)
    def dns(self) -> list[str]:
        return self.network_manager.get_config_value("dns")

    @export(property=True)
    def dhcp(self) -> list[bool]:
        return self.network_manager.get_config_value("dhcp")

    @export(property=True)
    def gateway(self) -> list[str]:
        return self.network_manager.get_config_value("gateway")

    @export(property=True)
    def interface(self) -> list[str]:
        return self.network_manager.get_config_value("interface")

    @export(property=True)
    def netmask(self) -> list[str]:
        return self.network_manager.get_config_value("netmask")

    @export(property=True)
    def version(self) -> str | None:
        distrib_description = self._os_release.get("DISTRIB_DESCRIPTION", "")
        name = self._os_release.get("NAME", "") or self._os_release.get("DISTRIB_ID", "")
        version = (
            self._os_release.get("VERSION", "")
            or self._os_release.get("VERSION_ID", "")
            or self._os_release.get("DISTRIB_RELEASE", "")
        )

        if not any([name, version, distrib_description]):
            return None

        if len(f"{name} {version}") > len(distrib_description):
            distrib_description = f"{name} {version}"

        return distrib_description or None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.LINUX.value



from pathlib import Path
from unittest.mock import Mock, patch

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loaders.libvirt import LibvirtLoader
from tests.conftest import absolute_path


def test_libvirt_detection(target_bare: Target, tmp_path: Path) -> None:
    xml_file = tmp_path.joinpath("base")
    xml_file.write_text("not a libvirt file")

    assert LibvirtLoader.detect(xml_file) is False

    xml_file = tmp_path.joinpath("base.xml")
    xml_file.write_text("<domain>")
    assert LibvirtLoader.detect(xml_file) is False

    xml_file = tmp_path.joinpath("base.xml")
    xml_file.write_text("not a libvirt file")

    assert LibvirtLoader.detect(xml_file) is False

    qemu_xml = Path(absolute_path("_data/loaders/libvirt/qemu.xml"))

    assert LibvirtLoader.detect(qemu_xml)


@patch(f"{LibvirtLoader.__module__}.container.open")
def test_libvirt_map(mocked_container_open: Mock, target_bare: Target, fs_linux: VirtualFilesystem) -> None:
    fs_linux.map_file("/test.xml", absolute_path("_data/loaders/libvirt/qemu.xml"))
    fs_linux.map_file_fh("/var/lib/libvirt/images/linux2022.qcow2", None)
    fs_linux.map_file_fh("/second-disk.qcow2", None)
    target_bare.fs.mount("/", fs_linux)

    loader = LibvirtLoader(target_bare.fs.path("/test.xml"))
    loader.map(target_bare)

    assert mocked_container_open.call_count == 2



import re
from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

re_firewall = re.compile(r"(.*)=(.*)")


class FirewallPlugin(Plugin):
    """Plugin that parses firewall rules from the registry."""

    KEY = "HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules"
    FIELD_MAP = {"app": "path"}
    VALUE_MAP = {"active": lambda val: val == "TRUE"}

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(self.KEY))) > 0:
            raise UnsupportedPluginError(f"Registry key {self.KEY} not found")

    @export(record=DynamicDescriptor(["uri"]))
    def firewall(self) -> Iterator[DynamicDescriptor]:
        """Return firewall rules saved in the registry.

        For a Windows operating system, the Firewall rules are stored in the
        HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules registry key.

        Yields dynamic records with usually the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            key (string): The rule key name.
            version (string): The version field of the rule.
            action (string): The action of the rule.
            active (boolean): Whether the rule is active.
            dir (string): The direction of the rule.
            protocol (uint32): The specified protocol (UDP=17, TCP=6).
            lport (string): The listening port of the rule.
            rport (string): The receiving port of the rule.
            profile (string): The Profile field of the rule.
            app (string): The App field of the rule.
            svc (string): The Svc of the rule.
            name (string): The Name of the rule.
            desc (string): The Desc of the rule.
            embed_ctxt (string): The EmbedCtxt of the rule.
        """
        for reg in self.target.registry.keys(self.KEY):
            for entry in reg.values():
                r = [
                    ("string", "key"),
                    ("string", "version"),
                ]
                data = {}

                fields = entry.value.split("|")
                version = fields[0]
                for field in fields[1:-1]:
                    fname, value = re_firewall.search(field).groups()
                    fname = fname.lower()

                    ft = self.FIELD_MAP.get(fname, "string")
                    try:
                        value = self.VALUE_MAP[fname](value)
                    except Exception:  # noqa
                        pass

                    r.append((ft, fname))
                    data[fname] = value

                if "app" in data:
                    data["app"] = self.target.resolve(data["app"])

                yield TargetRecordDescriptor("windows/registry/firewall", r)(
                    key=entry.name,
                    version=version,
                    _target=self.target,
                    **data,
                )



from __future__ import annotations

from typing import Iterator, Optional

from dissect.target.filesystem import Filesystem
from dissect.target.helpers import configutil
from dissect.target.helpers.record import EmptyRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.target import Target


class AndroidPlugin(LinuxPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self.target = target

        self.props = {}
        if (build_prop := self.target.fs.path("/build.prop")).exists():
            self.props = configutil.parse(build_prop, separator=("=",), comment_prefixes=("#",)).parsed_data

    @classmethod
    def detect(cls, target: Target) -> Optional[Filesystem]:
        for fs in target.filesystems:
            if fs.exists("/build.prop"):
                return fs
        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> AndroidPlugin:
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(property=True)
    def hostname(self) -> Optional[str]:
        return self.props.get("ro.build.host")

    @export(property=True)
    def ips(self) -> list[str]:
        return []

    @export(property=True)
    def version(self) -> str:
        full_version = "Android"

        release_version = self.props.get("ro.build.version.release")
        if release_version := self.props.get("ro.build.version.release"):
            full_version += f" {release_version}"

        if security_patch_version := self.props.get("ro.build.version.security_patch"):
            full_version += f" ({security_patch_version})"

        return full_version

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.ANDROID.value

    @export(record=EmptyRecord)
    def users(self) -> Iterator[EmptyRecord]:
        yield from ()



from dissect.target import Target
from dissect.target.plugins.filesystem.acquire_hash import AcquireHashPlugin
from tests._utils import absolute_path


def test_acquire_hash_plugin():
    file_hashes_target = Target().open(absolute_path("_data/plugins/filesystem/acquire_hash/test-acquire-hash.tar"))
    file_hashes_target.add_plugin(AcquireHashPlugin)

    results = list(file_hashes_target.acquire_hashes())

    assert results[0].path == "/sysvol/Windows/bfsvc.exe"
    assert len(results) == 998



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import logging

from dissect.target import Target
from dissect.target.exceptions import TargetError
from dissect.target.plugins.filesystem.yara import HAS_YARA, YaraPlugin
from dissect.target.tools.query import record_output
from dissect.target.tools.utils import (
    args_to_uri,
    catch_sigpipe,
    configure_generic_arguments,
    process_generic_arguments,
)

log = logging.getLogger(__name__)


@catch_sigpipe
def main():
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="target-yara",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )

    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="Targets to load")
    parser.add_argument("-s", "--strings", default=False, action="store_true", help="print output as string")
    parser.add_argument("-L", "--loader", action="store", default=None, help="select a specific loader (i.e. vmx, raw)")
    parser.add_argument("--children", action="store_true", help="include children")

    for args, kwargs in getattr(YaraPlugin.yara, "__args__", []):
        parser.add_argument(*args, **kwargs)

    configure_generic_arguments(parser)

    args, rest = parser.parse_known_args()
    process_generic_arguments(args)

    if not HAS_YARA:
        log.error("yara-python is not installed: pip install yara-python")
        parser.exit(1)

    if not args.targets:
        log.error("No targets provided")
        parser.exit(1)

    targets = args_to_uri(args.targets, args.loader, rest) if args.loader else args.targets

    try:
        for target in Target.open_all(targets, args.children):
            rs = record_output(args.strings, False)
            for record in target.yara(args.rules, args.path, args.max_size, args.check):
                rs.write(record)

    except TargetError as e:
        log.error(e)
        log.debug("", exc_info=e)
        parser.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import fnmatch
import re
from pathlib import Path
from typing import Any, BinaryIO, Iterator

from dissect.eventlog import evt
from flow.record import Record

from dissect.target import plugin
from dissect.target.exceptions import (
    FilesystemError,
    PluginError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
    UnsupportedPluginError,
)
from dissect.target.helpers.record import TargetRecordDescriptor

re_illegal_characters = re.compile(r"[\(\): \.\-#]")

EvtRecordDescriptor = TargetRecordDescriptor(
    "filesystem/windows/evt",
    [
        ("datetime", "ts"),
        ("datetime", "TimeGenerated"),
        ("datetime", "TimeWritten"),
        ("string", "SourceName"),
        ("varint", "EventID"),
        ("varint", "EventCode"),
        ("varint", "EventFacility"),
        ("varint", "EventCustomerFlag"),
        ("varint", "EventSeverity"),
        ("varint", "EventType"),
        ("varint", "EventCategory"),
        ("string", "Computername"),
        ("string", "UserSid"),
        ("string[]", "Strings"),
        ("bytes", "Data"),
    ],
)


EVT_GLOB = "*.evt"


class WindowsEventlogsMixin:
    EVENTLOG_REGISTRY_KEY = "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Eventlog"
    LOGS_DIR_PATH = None

    @plugin.internal
    def get_logs(self, filename_glob="*") -> list[Path]:
        file_paths = []
        file_paths.extend(self.get_logs_from_dir(self.LOGS_DIR_PATH, filename_glob=filename_glob))

        if self.EVENTLOG_REGISTRY_KEY:
            for reg_path in self.get_logs_from_registry(filename_glob=filename_glob):
                # We can't filter duplicates on file path alone, since "sysvol" and "C:" can show up interchangeably.
                try:
                    if any(fpath.samefile(reg_path) for fpath in file_paths):
                        continue
                except FilesystemError:
                    pass

                file_paths.append(reg_path)

        return file_paths

    @plugin.internal
    def get_logs_from_dir(self, logs_dir: str, filename_glob: str = "*") -> list[Path]:
        file_paths = []
        logs_dir = self.target.fs.path(logs_dir)
        if logs_dir.exists():
            file_paths.extend(list(logs_dir.glob(filename_glob)))

        self.target.log.debug("Log files found in '%s': %d", self.LOGS_DIR_PATH, len(file_paths))
        return file_paths

    @plugin.internal
    def get_logs_from_registry(self, filename_glob: str = "*") -> list[Path]:
        # compile glob into case-insensitive regex
        filename_regex = re.compile(fnmatch.translate(filename_glob), re.IGNORECASE)

        file_paths = []

        try:
            subkeys = self.target.registry.key(self.EVENTLOG_REGISTRY_KEY).subkeys()
        except RegistryKeyNotFoundError:
            self.target.log.warning("No eventlog registry key %s found", self.EVENTLOG_REGISTRY_KEY)
            return []
        except PluginError:
            self.target.log.warning("Cannot access registry in target")
            return []

        for subkey in subkeys:
            try:
                subkey_value = subkey.value("File")
            except RegistryValueNotFoundError:
                continue
            file_paths.append(subkey_value.value)

        # resolve aliases (like `%systemroot%`) in the paths
        file_paths = [self.target.resolve(p) for p in file_paths]
        file_paths = [path for path in file_paths if filename_regex.match(str(path))]

        self.target.log.debug("Log files found in '%s': %d", self.EVENTLOG_REGISTRY_KEY, len(file_paths))

        return file_paths

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.LOGS_DIR_PATH).exists():
            raise UnsupportedPluginError(f'Event log directory "{self.LOGS_DIR_PATH}" not found')


class EvtPlugin(WindowsEventlogsMixin, plugin.Plugin):
    """Windows ``.evt`` event log plugin."""

    LOGS_DIR_PATH = "sysvol/windows/system32/config"

    NEEDLE = b"LfLe"
    CHUNK_SIZE = 0x10000

    @plugin.arg("--logs-dir", help="logs directory to scan")
    @plugin.arg("--log-file-glob", default=EVT_GLOB, help="glob pattern to match a log file name")
    @plugin.export(record=EvtRecordDescriptor)
    def evt(self, log_file_glob: str = EVT_GLOB, logs_dir: str | None = None) -> Iterator[EvtRecordDescriptor]:
        """Parse Windows Eventlog files (``*.evt``).

        Yields dynamically created records based on the fields in the event.
        At least contains the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The TimeCreated_SystemTime field of the event.
            Provider_Name (string): The Provider_Name field of the event.
            EventID (int): The EventID of the event.
        """

        if logs_dir:
            log_paths = self.get_logs_from_dir(logs_dir, filename_glob=log_file_glob)
        else:
            log_paths = self.get_logs(filename_glob=log_file_glob)

        for entry in log_paths:
            if not entry.exists():
                self.target.log.warning("Event log file does not exist: %s", entry)
                continue

            try:
                entry_data = entry.open()
            except FilesystemError:
                self.target.log.exception("Failed to open event log: %s", entry)
                continue

            for record in evt.Evt(entry_data):
                yield self._build_record(record)

    def _build_record(self, record: Any) -> Record:
        return EvtRecordDescriptor(
            ts=record.TimeGenerated,
            TimeGenerated=record.TimeGenerated,
            TimeWritten=record.TimeWritten,
            SourceName=record.SourceName,
            EventID=record.EventID,
            EventCode=record.EventCode,
            EventFacility=record.EventFacility,
            EventCustomerFlag=record.EventCustomerFlag,
            EventSeverity=record.EventSeverity,
            EventType=record.EventType,
            EventCategory=record.EventCategory,
            Computername=record.Computername,
            Strings=record.Strings,
            Data=record.Data,
            _target=self.target,
        )

    @plugin.export(record=EvtRecordDescriptor)
    def scraped_evt(self) -> Iterator[EvtRecordDescriptor]:
        """Yields EVT log file records scraped from target disks"""
        yield from self.target.scrape.scrape_chunks_from_disks(
            needle=self.NEEDLE,
            chunk_size=self.CHUNK_SIZE,
            chunk_parser=self._parse_chunk,
            chunk_reader=self._read_chunk,
        )

    def _read_chunk(self, fh: BinaryIO, needle: bytes, offset: int, chunk_size: int) -> bytes:
        # Needle is a 2nd field in EVT header, so we need to back up
        # by 4 bytes to get to the start of the the record.
        fh.seek(offset - 4)
        return fh.read(chunk_size)

    def _parse_chunk(self, _, chunk: bytes) -> Iterator[Record]:
        for record in evt.parse_chunk(chunk):
            yield self._build_record(record)



import struct
from io import BytesIO
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from dissect.target import container
from dissect.target.containers import raw, vhd
from dissect.target.exceptions import ContainerError


@pytest.fixture
def mocked_ewf_detect():
    mocked_ewf = Mock()
    mocked_ewf.EwfContainer.detect.return_value = True
    mocked_ewf.EwfContainer.detect
    with patch.object(container, "CONTAINERS", [mocked_ewf.EwfContainer]):
        yield mocked_ewf.EwfContainer


@pytest.mark.parametrize(
    "path, expected_output",
    [
        ("hello", Path("hello")),
        (["hello"], [Path("hello")]),
        ([Path("hello")], [Path("hello")]),
    ],
)
def test_open_inputs(mocked_ewf_detect: Mock, path, expected_output):
    container.open(path)
    mocked_ewf_detect.assert_called_with(expected_output)


def test_open_fallback_fh(tmp_path):
    # Create a valid VHD file
    fake_vhd = (
        (bytes(range(256)) * 2)
        + b"conectix"
        + (b"\x00" * 8)
        + (b"\xFF" * 8)
        + (b"\x00" * 24)
        + struct.pack(">Q", 512)
        + (b"\x00" * 455)
    )

    tmp_with_ext = tmp_path.joinpath("testfile.vhd")
    tmp_without_ext = tmp_path.joinpath("testfile")
    tmp_with_wrong_ext = tmp_path.joinpath("testfile.qcow2")

    for path in [tmp_with_ext, tmp_without_ext, tmp_with_wrong_ext]:
        path.write_bytes(fake_vhd)

        assert isinstance(container.open(path), vhd.VhdContainer)
        assert vhd.VhdContainer.detect(path)

        with path.open("rb") as fh:
            assert isinstance(container.open(fh), vhd.VhdContainer)

        with path.open("rb") as fh:
            assert vhd.VhdContainer.detect(fh)

    tmp_nonexistent = tmp_path.joinpath("doesntexist")
    with pytest.raises(ContainerError):
        container.open(tmp_nonexistent)

    assert not vhd.VhdContainer.detect(tmp_nonexistent)

    tmp_dummy = tmp_path.joinpath("testdummy")
    tmp_dummy.write_bytes(b"\x00" * 1024)
    assert isinstance(container.open(tmp_dummy), raw.RawContainer)
    assert not vhd.VhdContainer.detect(tmp_dummy)


def test_reset_file_position() -> None:
    fh = BytesIO(b"\x00" * 8192)
    fh.seek(512)

    class MockContainer(container.Container):
        def __init__(self, fh):
            assert fh.tell() == 0
            fh.seek(1024)
            self.success = True

        @staticmethod
        def _detect_fh(fh, *args, **kwargs):
            assert fh.tell() == 0
            fh.seek(256)
            return True

    mock_container = Mock()
    mock_container.MockContainer = MockContainer
    with patch.object(container, "CONTAINERS", [mock_container.MockContainer]):
        opened_container = container.open(fh)
        assert isinstance(opened_container, mock_container.MockContainer)
        assert opened_container.success
        assert fh.tell() == 512



from __future__ import annotations

from functools import lru_cache
from typing import Iterator, NamedTuple, Union

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import UnixUserRecord, WindowsUserRecord
from dissect.target.plugin import InternalPlugin

UserRecord = Union[UnixUserRecord, WindowsUserRecord]


class UserDetails(NamedTuple):
    user: UserRecord
    home_path: TargetPath | None


class UsersPlugin(InternalPlugin):
    """Internal plugin that provides helper functions for retrieving user details."""

    __namespace__ = "user_details"

    def __init__(self, target: Target):
        super().__init__(target)
        self.find = lru_cache(32)(self.find)

    def check_compatible(self) -> None:
        if not hasattr(self.target, "users"):
            raise UnsupportedPluginError("Unsupported Plugin")

    def find(
        self,
        sid: str | None = None,
        uid: str | None = None,
        username: str | None = None,
        force_case_sensitive: bool = False,
    ) -> UserDetails | None:
        """Find User record matching provided sid, uid or username and return UserDetails object"""
        if all(map(lambda x: x is None, [sid, uid, username])):
            raise ValueError("Either sid or uid or username is expected")

        def is_name_matching(name: str) -> bool:
            if force_case_sensitive or self.target.os != "windows":
                # always do case-sensitive match for non-Windows OSes
                return name == username
            else:
                return name.lower() == username.lower()

        for user in self.target.users():
            if (
                (sid is not None and user.sid == sid)
                or (uid is not None and user.uid == uid)
                or (username is not None and is_name_matching(user.name))
            ):
                return self.get(user)

    def get(self, user: UserRecord) -> UserDetails:
        """Return additional details about the user"""
        # Resolving the user home can not use the user's environment variables,
        # as those depend on the user's home to be known first. So we resolve
        # without providing the user (s)id.
        home_path = self.target.fs.path(self.target.resolve(str(user.home))) if user.home else None
        return UserDetails(user=user, home_path=home_path)

    def all(self) -> Iterator[UserDetails]:
        """Return UserDetails objects for all users found"""
        for user in self.target.users():
            yield self.get(user)

    def all_with_home(self) -> Iterator[UserDetails]:
        """Return UserDetails objects for users that have existing directory set as home directory"""
        for user in self.target.users():
            if user.home:
                user_details = self.get(user)
                if user_details.home_path.exists():
                    yield user_details



import gzip
import tarfile
import textwrap
from datetime import datetime, timezone
from io import BytesIO
from unittest.mock import patch
from zoneinfo import ZoneInfo

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.plugins.general import default
from dissect.target.plugins.os.unix.log.messages import MessagesPlugin, MessagesRecord
from tests._utils import absolute_path


def test_unix_log_messages_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))

    data_file = absolute_path("_data/plugins/os/unix/log/messages/messages")
    fs_unix.map_file("var/log/messages", data_file)

    entry = fs_unix.get("var/log/messages")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix_users.add_plugin(MessagesPlugin)

        results = list(target_unix_users.messages())
        assert len(results) == 2

        assert results[0].ts == datetime(2022, 1, 1, 13, 21, 34, tzinfo=ZoneInfo("Europe/Amsterdam"))
        assert results[0].message == "Stopped target Swap."
        assert results[0].pid is None
        assert results[0].source == "/var/log/messages"

        assert results[1].ts == datetime(2021, 12, 31, 3, 14, 15, tzinfo=ZoneInfo("Europe/Amsterdam"))
        assert results[1].message == "Starting Journal Service..."
        assert results[1].pid == 1
        assert results[1].source == "/var/log/messages"

    # assure syslog() behaves the same as messages()
    syslogs = list(target_unix_users.syslog())
    assert len(syslogs) == len(results)
    assert isinstance(syslogs[0], type(MessagesRecord()))


def test_unix_log_messages_compressed_timezone_year_rollover() -> None:
    target = Target()
    bio = BytesIO()

    with tarfile.open(mode="w:gz", fileobj=bio) as tf:
        # UTC-6 tar filesystem
        timezone_contents = "America/Chicago"
        timezone_file = BytesIO(textwrap.dedent(timezone_contents).encode())
        timezone_tar_info = tarfile.TarInfo("etc/timezone")
        timezone_tar_info.size = len(timezone_contents)
        tf.addfile(timezone_tar_info, timezone_file)

        # Create a tar file with correct mtime
        messages_log_contents = """\
        Dec 31 03:14:00 localhost systemd[1]: Starting Journal Service...
        \xfean  1 03:14:00 localhost systemd[1]: Starting Journal Service...
        Jan  1 13:37:00 localhost systemd: Stopped target Swap."""
        messages_log_file = BytesIO(textwrap.dedent(messages_log_contents).encode())
        messages_log_tar_info = tarfile.TarInfo("var/log/messages")
        messages_log_tar_info.size = len(textwrap.dedent(messages_log_contents))
        messages_log_tar_info.mtime = 1640995800  # (Sat 1 January 2022 00:10:00 UTC)
        tf.addfile(messages_log_tar_info, messages_log_file)

    fs = TarFilesystem(bio)
    target.filesystems.add(fs)
    target.fs.mount("/", fs)
    target._os_plugin = default.DefaultPlugin
    target.apply()
    target.add_plugin(MessagesPlugin)

    results = list(target.messages())
    results.reverse()

    assert len(results) == 2
    assert isinstance(results[0], type(MessagesRecord()))
    assert isinstance(results[1], type(MessagesRecord()))
    assert results[0].ts == datetime(2020, 12, 31, 3, 14, 0, tzinfo=ZoneInfo("America/Chicago"))
    assert results[1].ts == datetime(2021, 1, 1, 13, 37, 0, tzinfo=ZoneInfo("America/Chicago"))


def test_unix_log_messages_malformed_log_year_rollover(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))

    messages = BytesIO(
        b"Dec 31 03:14:00 localhost systemd[1]: Starting Journal Service...\r\n"
        b"\xefan  1 13:37:00 localhost systemd: Stopped target Swap.\r\n"
        b"Dec 31 03:14:00 localhost systemd[1]: Starting Journal Service...\r\n"
    )
    fs_unix.map_file_fh("var/log/messages", messages)

    entry = fs_unix.get("var/log/messages")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix_users.add_plugin(MessagesPlugin)

        results = list(target_unix_users.messages())
        assert len(results) == 2


def test_unix_messages_cloud_init(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """test if we correctly parse plaintext and compressed cloud-init log files."""

    messages = """
    2005-08-09 11:55:21,000 - foo.py[DEBUG]: This is a cloud-init message!
    2005-08-09 11:55:21,001 - util.py[DEBUG]: Cloud-init v. 1.2.3-4ubuntu5 running 'init-local' at Tue, 9 Aug 2005 11:55:21 +0000. Up 13.37 seconds.
    """  # noqa: E501
    msg_bytes = textwrap.dedent(messages).encode()

    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))
    fs_unix.map_file_fh("/var/log/installer/cloud-init.log", BytesIO(msg_bytes))
    fs_unix.map_file_fh("/var/log/installer/cloud-init.log.1.gz", BytesIO(gzip.compress(msg_bytes)))
    target_unix.add_plugin(MessagesPlugin)

    results = sorted(list(target_unix.messages()), key=lambda r: r.source)
    assert len(results) == 4

    assert results[0].ts == datetime(2005, 8, 9, 11, 55, 21, 0, tzinfo=ZoneInfo("Europe/Amsterdam"))
    assert results[0].daemon == "foo.py"
    assert results[0].pid is None
    assert results[0].message == "This is a cloud-init message!"
    assert results[0].source == "/var/log/installer/cloud-init.log"

    assert results[-1].ts == datetime(2005, 8, 9, 11, 55, 21, 1_000, tzinfo=ZoneInfo("Europe/Amsterdam"))
    assert results[-1].daemon == "util.py"
    assert results[-1].pid is None
    assert (
        results[-1].message
        == "Cloud-init v. 1.2.3-4ubuntu5 running 'init-local' at Tue, 9 Aug 2005 11:55:21 +0000. Up 13.37 seconds."  # noqa: E501
    )
    assert results[-1].source == "/var/log/installer/cloud-init.log.1.gz"



import calendar
from collections import namedtuple
from datetime import datetime, timedelta, timezone, tzinfo
from typing import Dict, Tuple

from dissect.cstruct import cstruct

from dissect.target.exceptions import (
    RegistryError,
    RegistryValueNotFoundError,
    UnsupportedPluginError,
)
from dissect.target.helpers.mui import MUI_TZ_MAP
from dissect.target.helpers.regutil import RegistryKey
from dissect.target.plugin import Plugin, internal

tz_def = """
typedef struct _SYSTEMTIME {
    WORD wYear;
    WORD wMonth;
    WORD wDayOfWeek;
    WORD wDay;
    WORD wHour;
    WORD wMinute;
    WORD wSecond;
    WORD wMilliseconds;
} SYSTEMTIME;

typedef struct _REG_TZI_FORMAT {
    LONG Bias;
    LONG StandardBias;
    LONG DaylightBias;
    SYSTEMTIME StandardDate;
    SYSTEMTIME DaylightDate;
} REG_TZI_FORMAT;
"""
c_tz = cstruct().load(tz_def)


# Althoug calendar.SUNDAY is only officially documented since Python 3.10, it
# is present in Python 3.9, so we ignore the vermin warnings.
SUNDAY = calendar.SUNDAY  # novermin
SundayFirstCalendar = calendar.Calendar(SUNDAY)
TimezoneInformation = namedtuple(
    "TimezoneInformation",
    (
        # The current bias from UTC
        "bias",
        # The additional bias on top of the current bias when in standard time (usually 0)
        "standard_bias",
        # The additional bias on top of the current bias when in daylight saving time (usually -60)
        "daylight_bias",
        # When standard time starts, or in other words, daylight saving ends (see parse_systemtime_transition)
        "standard_date",
        # When daylight saving starts (see parse_systemtime_transition)
        "daylight_date",
    ),
)


ZERO = timedelta(0)
HOUR = timedelta(hours=1)


def parse_systemtime_transition(systemtime: c_tz._SYSTEMTIME, year: int) -> datetime:
    """Return the transition datetime for a given year using the SYSTEMTIME of a STD or DST transition date.

    The SYSTEMTIME date of a TZI structure needs to be used to calculate the actual date for a given year.
    The wMonth member indicates the month, the wDayOfWeek member indicates the weekday and the wDay indicates the
    occurance of the day of the week within the month (1 to 5, where 5 indicates the final occurrence during the
    month if that day of the week does not occur 5 times).

    Reference:
        - https://docs.microsoft.com/en-us/windows/win32/api/timezoneapi/ns-timezoneapi-time_zone_information
    """
    month = SundayFirstCalendar.monthdayscalendar(year, systemtime.wMonth)
    occurrences = [week[systemtime.wDayOfWeek] for week in month if week[systemtime.wDayOfWeek]]
    target_occurrence = -1 if systemtime.wDay == 5 and len(occurrences) < 5 else systemtime.wDay - 1

    return datetime(
        year,
        systemtime.wMonth,
        occurrences[target_occurrence],
        systemtime.wHour,
        systemtime.wMinute,
        systemtime.wSecond,
        systemtime.wMilliseconds * 1000,
    )


def parse_dynamic_dst(key: RegistryKey) -> Dict[int, TimezoneInformation]:
    """Parse dynamic DST information from a given TimeZoneInformation registry key.

    If a timezone has dynamic DST information, there's a "Dynamic DST" subkey with values for each year.
    The FirstEntry and LastEntry contain the first and last year respectively. The TZI structure is the same as the
    main TimeZoneInformation TZI.
    """
    result = {}
    try:
        dynamic_key = key.subkey("Dynamic DST")
        first = int(dynamic_key.value("FirstEntry").value)
        last = int(dynamic_key.value("LastEntry").value)

        for year in range(first, last + 1):
            result[year] = parse_tzi(dynamic_key.value(str(year)).value)
    except RegistryError:
        pass

    return result


def parse_tzi(tzi: bytes) -> TimezoneInformation:
    """Parse binary TZI into a TimezoneInformation namedtuple."""
    parsed = c_tz.REG_TZI_FORMAT(tzi)
    return TimezoneInformation(
        parsed.Bias,
        parsed.StandardBias,
        parsed.DaylightBias,
        parsed.StandardDate,
        parsed.DaylightDate,
    )


def get_dst_range(tzi: TimezoneInformation, year: int) -> Tuple[datetime, datetime]:
    """Get the start and end date of DST for the given year."""
    start = parse_systemtime_transition(tzi.daylight_date, year)
    end = parse_systemtime_transition(tzi.standard_date, year)

    return (start, end)


class WindowsTimezone(tzinfo):
    """A datetime.tzinfo class representing a timezone from parsed Windows TZI data.

    Mostly inspired by the examples in the Python documentation.
    """

    def __init__(self, name: str, key: RegistryKey):
        self.name = name
        self.display = translate_tz(key, "Display")
        self.dlt_name = translate_tz(key, "Dlt")
        self.std_name = translate_tz(key, "Std")
        self.tzi = parse_tzi(key.value("TZI").value)
        self.dynamic_dst = parse_dynamic_dst(key)

        self.bias = timedelta(minutes=-self.tzi.bias)

    def __repr__(self) -> str:
        return self.display

    def is_dst(self, dt: datetime) -> bool:
        assert dt.tzinfo is self

        tzi = self.dynamic_dst.get(dt.year, self.tzi)
        start, end = get_dst_range(tzi, dt.year)

        # DST is flipped in some regions
        flip = False
        if start > end:
            flip = True
            start, end = end, start

        # Can't compare naive to aware objects, so strip the timezone from
        # dt first.
        dt = dt.replace(tzinfo=None)

        result = False
        if start + HOUR <= dt < end - HOUR:
            # DST is in effect.
            result = True
        elif end - HOUR <= dt < end:
            # Fold (an ambiguous hour): use dt.fold to disambiguate.
            result = False if dt.fold else True
        elif start <= dt < start + HOUR:
            # Gap (a non-existent hour): reverse the fold rule.
            result = True if dt.fold else False
        else:
            # DST is off.
            result = False

        return not result if flip else result

    def utcoffset(self, dt: datetime) -> int:
        # Windows timezones use a default bias and a separate additional bias when in STD or DST
        return self.bias + self.dst(dt)

    def dst(self, dt: datetime) -> timedelta:
        tzi = self.dynamic_dst.get(dt.year, self.tzi)
        std_bias = timedelta(minutes=-tzi.standard_bias)
        dst_bias = timedelta(minutes=-tzi.daylight_bias)

        return dst_bias if self.is_dst(dt) else std_bias

    def tzname(self, dt: datetime) -> str:
        return self.dlt_name if self.is_dst(dt) else self.std_name


class DateTimePlugin(Plugin):
    __namespace__ = "datetime"

    def __init__(self, target):
        super().__init__(target)

        tz_name = None

        timezone_information = self.target.registry.key("HKLM\\SYSTEM\\CurrentControlSet\\Control\\TimeZoneInformation")
        try:
            # Windows 7+
            tz_name = timezone_information.value("TimeZoneKeyName").value
        except RegistryValueNotFoundError:
            # < Windows 7 (https://github.com/dateutil/dateutil/issues/210)
            tz_name_localized = timezone_information.value("StandardName").value
            for timezone_key in self.target.registry.key(
                "HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones"
            ).subkeys():
                if tz_name_localized == timezone_key.value("Std").value:
                    tz_name = timezone_key.name

        if tz_name is None:
            self.target.log.error("Failed to load timezone information")
            self._tzinfo = None
        else:
            self._tzinfo = self.tz(tz_name)

    def check_compatible(self) -> None:
        if not self._tzinfo:
            raise UnsupportedPluginError("No time zone information")

    @internal
    def tz(self, name: str) -> tzinfo:
        """Return a datetime.tzinfo of the given timezone name."""
        tz_data = self.target.registry.key(f"HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\{name}")
        return WindowsTimezone(name, tz_data)

    @internal(property=True)
    def tzinfo(self) -> tzinfo:
        """Return a datetime.tzinfo of the current system timezone."""
        return self._tzinfo

    @internal
    def local(self, dt: datetime) -> datetime:
        """Replace the tzinfo of a given datetime.datetime object with the current system tzinfo without conversion."""
        return dt.replace(tzinfo=self._tzinfo)

    @internal
    def to_utc(self, dt: datetime) -> datetime:
        """Convert any datetime.datetime object into a UTC datetime.datetime object.

        First replaces the current tzinfo with the system tzinfo without conversion, then converts it to an aware
        UTC datetime object.
        """
        return self.local(dt).astimezone(timezone.utc)


def translate_tz(key: RegistryKey, name: str) -> str:
    """Translate a timezone resource string to English.

    Non-English distributions of Windows contain a local translation in the "Display", "Dlt" and "Std" keys.
    The ``MUI_*`` keys contain a reference to the English timezone name we want, e.g. "@tzres.dll,-1337".
    """
    try:
        string_id = int(key.value("MUI_" + name).value.replace("@tzres.dll,-", ""))
        if translation := MUI_TZ_MAP.get(string_id):
            return translation
    except (RegistryValueNotFoundError, ValueError):
        pass

    return key.value(name).value



from __future__ import annotations

import fnmatch
import re
from pathlib import Path
from typing import Iterator

from dissect.target import Target
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import arg, export
from dissect.target.plugins.general.config import (
    ConfigurationEntry,
    ConfigurationTreePlugin,
)

UnixConfigTreeRecord = TargetRecordDescriptor(
    "unix/config",
    [
        ("path", "source"),
        ("path", "path"),
        ("string", "key"),
        ("string[]", "value"),
    ],
)


class EtcTree(ConfigurationTreePlugin):
    """Unix etc configuration tree plugin."""

    __namespace__ = "etc"

    def __init__(self, target: Target):
        super().__init__(target, "/etc")

    def _sub(
        self, items: ConfigurationEntry | dict, entry: Path, orig_path: Path, pattern: str
    ) -> Iterator[UnixConfigTreeRecord]:
        index = 0
        if not isinstance(items, dict):
            items = items.as_dict()

        for raw_key, value in items.items():
            key = re.sub(r"[\n\r\t]", "", raw_key)
            path = Path(entry) / Path(key)

            if isinstance(value, dict):
                yield from self._sub(value, path, orig_path, pattern)
                continue

            if not isinstance(value, list):
                value = [str(value)]

            if fnmatch.fnmatch(path, pattern):
                data = {
                    "_target": self.target,
                    "source": self.target.fs.path(orig_path),
                    "path": path,
                    "key": key,
                    "value": value,
                }
                if value == [""]:
                    data["key"] = index
                    data["value"] = [key]
                    index += 1

                yield UnixConfigTreeRecord(**data)

    @export(record=UnixConfigTreeRecord)
    @arg("--glob", dest="pattern", required=False, default="*", type=str, help="Glob-style pattern to search for")
    @arg("--root", dest="root", required=False, default="/", type=str, help="Path to use as root for search")
    def etc(self, pattern: str, root: str) -> Iterator[UnixConfigTreeRecord]:
        """Yield etc configuration records."""

        for entry, subs, items in self.config_fs.walk(root):
            for item in items:
                try:
                    path = Path(entry) / item
                    config_object = self.get(str(path))

                    if isinstance(config_object, ConfigurationEntry):
                        yield from self._sub(config_object, path, orig_path=path, pattern=pattern)
                except Exception:
                    self.target.log.warning("Could not open configuration item: %s", item)
                    pass



from __future__ import annotations

import logging
import warnings
from typing import Iterator, Union

from flow.record import GroupedRecord

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.task_helpers.tasks_job import AtTask
from dissect.target.plugins.os.windows.task_helpers.tasks_xml import ScheduledTasks

warnings.simplefilter(action="ignore", category=FutureWarning)
log = logging.getLogger(__name__)

TaskRecord = TargetRecordDescriptor(
    "filesystem/windows/task",
    [
        ("path", "task_path"),
        ("string", "uri"),
        ("string", "security_descriptor"),
        ("string", "source"),
        ("datetime", "date"),
        ("datetime", "last_run_date"),
        ("string", "author"),
        ("string", "version"),
        ("string", "description"),
        ("string", "documentation"),
        ("string", "task_name"),
        ("string", "app_name"),
        ("string", "args"),
        ("string", "start_in"),
        ("string", "comment"),
        ("string", "run_as"),
        ("string", "cpassword"),
        ("string", "enabled"),
        ("string", "action"),
        ("string", "principal_id"),
        ("string", "user_id"),
        ("string", "logon_type"),
        ("string", "group_id"),
        ("string", "display_name"),
        ("string", "run_level"),
        ("string", "process_token_sid_type"),
        ("string", "required_privileges"),
        ("string", "restart_on_failure_interval"),
        ("string", "restart_on_failure_count"),
        ("string", "mutiple_instances_policy"),
        ("string", "dissalow_start_on_batteries"),
        ("string", "stop_going_on_batteries"),
        ("string", "allow_start_on_demand"),
        ("string", "start_when_available"),
        ("string", "network_profile_name"),
        ("string", "run_only_network_available"),
        ("string", "wake_to_run"),
        ("string", "enabled"),
        ("string", "hidden"),
        ("string", "delete_expired_task_after"),
        ("string", "idle_duration"),
        ("string", "idle_wait_timeout"),
        ("string", "idle_stop_on_idle_end"),
        ("string", "idle_restart_on_idle"),
        ("string", "network_settings_name"),
        ("string", "network_settings_id"),
        ("string", "execution_time_limit"),
        ("string", "priority"),
        ("string", "run_only_idle"),
        ("string", "unified_scheduling_engine"),
        ("string", "disallow_start_on_remote_app_session"),
        ("string", "data"),
        ("string", "raw_data"),
    ],
)


class TasksPlugin(Plugin):
    """Plugin for retrieving scheduled tasks on a Windows system.

    Args:
        target: The target system.
    """

    PATHS = {
        "sysvol/windows/system32/tasks",
        "sysvol/windows/system32/tasks_migrated",
        "sysvol/windows/syswow64/tasks",
        "sysvol/windows/tasks",  # at.exe job file location
    }
    GLOB_PATHS = [
        "sysvol/windows/system32/GroupPolicy/DataStore/*/Machine/Preferences/ScheduledTasks/*",
        "sysvol/ProgramData/Microsoft/*/Preferences/ScheduledTasks/*",
    ]

    def __init__(self, target: Target):
        super().__init__(target)
        self.task_files = []

        for path in self.GLOB_PATHS:
            start_path, pattern = path.split("*", 1)
            for entry in self.target.fs.path(start_path).rglob("*" + pattern):
                if entry.is_file() and entry.suffix == ".xml":
                    self.task_files.append(entry)

        for file_path in self.PATHS:
            fpath = self.target.fs.path(file_path)
            if not fpath.exists():
                continue

            for entry in fpath.rglob("*"):
                if entry.is_file() and (entry.suffix.lower() == ".job" or not entry.suffix):
                    self.task_files.append(entry)

    def check_compatible(self) -> None:
        if len(self.task_files) == 0:
            raise UnsupportedPluginError("No task files")

    @export(record=DynamicDescriptor(["path", "datetime"]))
    def tasks(self) -> Iterator[Union[TaskRecord, GroupedRecord]]:
        """Return all scheduled tasks on a Windows system.

        On a Windows system, a scheduled task is a program or script that is executed on a specific time or at specific
        intervals. An adversary may leverage such scheduled tasks to gain persistence on a system.

        References:
            - https://en.wikipedia.org/wiki/Windows_Task_Scheduler

        Yields:
            The scheduled tasks found on the target.
        """
        for task_file in self.task_files:
            if not task_file.suffix or task_file.suffix == ".xml":
                task_objects = ScheduledTasks(task_file).tasks
            else:
                task_objects = [AtTask(task_file, self.target)]

            for task_object in task_objects:
                record_kwargs = {}
                for attr in TaskRecord.fields.keys():
                    record_kwargs[attr] = getattr(task_object, attr, None)

                record = TaskRecord(**record_kwargs, _target=self.target)
                yield record

                # Actions
                for action in task_object.get_actions():
                    grouped = GroupedRecord("filesystem/windows/task/grouped", [record, action])
                    yield grouped

                # Triggers
                for trigger in task_object.get_triggers():
                    grouped = GroupedRecord("filesystem/windows/task/grouped", [record, trigger])
                    yield grouped



import plistlib

from dissect.target.helpers.record import EmptyRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin


class IOSPlugin(BsdPlugin):
    @classmethod
    def detect(cls, target):
        for fs in target.filesystems:
            if fs.exists("/private/var/preferences"):
                return fs

        return None

    @classmethod
    def create(cls, target, sysvol):
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(property=True)
    def hostname(self):
        path = self.target.fs.path("/private/var/preferences/SystemConfiguration/preferences.plist")

        if not path.exists():
            return None

        preferences = plistlib.load(path.open())
        return preferences["System"]["System"]["ComputerName"]

    @export(property=True)
    def ips(self):
        raise NotImplementedError

    @export(property=True)
    def version(self):
        raise NotImplementedError

    @export(record=EmptyRecord)
    def users(self):
        raise NotImplementedError

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.IOS.value



import importlib


class LazyImport:
    def __init__(self, module_name):
        self._module_name = module_name
        self._module = None

    def _import(self):
        if not self._module:
            try:
                self._module = importlib.import_module(self._module_name)
            except Exception as e:
                self._module = FailedImport(e, self)

    def __getattr__(self, attr):
        return LazyAttr(attr, self)

    def __repr__(self):
        return (
            f"<lazy {self._module_name} loaded={self._module is not None} "
            f"failed={isinstance(self._module, FailedImport)}>"
        )


class FailedImport:
    def __init__(self, exc, module):
        self.exc = exc
        self.module = module

    def _error(self):
        raise ImportError(f"Failed to lazily import {self.module._module_name}: {self.exc}")  # noqa

    def __call__(self, *args, **kwargs):
        self._error()

    def __getattr__(self, attr):
        self._error()


class LazyAttr:
    def __init__(self, attr, module):
        self.attr = attr
        self.module = module
        self._realattr = None

    @property
    def realattr(self):
        if not self._realattr:
            self.module._import()
            self._realattr = getattr(self.module._module, self.attr)

        return self._realattr

    @property
    def __doc__(self):
        return self.realattr.__doc__

    def __call__(self, *args, **kwargs):
        return self.realattr(*args, **kwargs)

    def __getattr__(self, attr):
        return getattr(self.realattr, attr)

    def __repr__(self):
        return f"<lazyattr {self.module._module_name}.{self.attr} loaded={self.realattr is not None}>"


def import_lazy(module_name):
    return LazyImport(module_name)



from __future__ import annotations

import tempfile
from io import BytesIO
from pathlib import Path
from typing import Iterator

import pytest

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.filesystem.yara import HAS_YARA, YaraPlugin, is_valid_yara
from tests._utils import absolute_path

if HAS_YARA:
    import yara

rule_file = absolute_path("_data/plugins/filesystem/yara/rule.yar")
another_rule_file = absolute_path("_data/plugins/filesystem/yara/another.yar")
invalid_rule = absolute_path("_data/plugins/filesystem/yara/invalid.yar")
rule_dir = Path(rule_file).parent


@pytest.fixture
def target_yara(target_default: Target) -> Iterator[Target]:
    vfs = VirtualFilesystem()
    vfs.map_file_fh("test_file", BytesIO(b"test string"))
    vfs.map_file_fh("/test/dir/to/test_file", BytesIO(b"test string"))
    vfs.map_file_fh("should_not_hit", BytesIO(b"this is another file."))
    target_default.fs.mount("/", vfs)
    target_default.add_plugin(YaraPlugin)
    yield target_default


@pytest.mark.skipif(not HAS_YARA, reason="requires python-yara")
def test_yara_plugin(target_yara: Target) -> None:
    results = list(target_yara.yara(rules=[Path(rule_file)]))

    assert len(results) == 2
    assert results[0].path == "/test_file"
    assert results[1].path == "/test/dir/to/test_file"
    assert results[0].rule == "test_rule_name"


@pytest.mark.skipif(not HAS_YARA, reason="requires python-yara")
@pytest.mark.parametrize(
    "rules,expected_hits,should_be_valid",
    [
        (["/does/not/exist"], 0, False),
        ([rule_file, rule_file], 2, True),
        ([rule_file, another_rule_file], 4, True),
        ([rule_dir], 4, False),  # contains invalid.yar
        ([invalid_rule], 0, False),
    ],
)
def test_yara_plugin_invalid_rules(
    target_yara: Target, rules: list[str | Path], expected_hits: int, should_be_valid: bool
) -> None:
    assert is_valid_yara(files={str(file): file for file in rules}) == should_be_valid

    results = list(target_yara.yara(rules=rules, check=True))
    assert len(results) == expected_hits


@pytest.mark.skipif(not HAS_YARA, reason="requires python-yara")
def test_yara_plugin_invalid_rule_warn(target_yara: Target, caplog: pytest.CaptureFixture) -> None:
    results = list(target_yara.yara(rules=[invalid_rule, another_rule_file], check=True))
    assert "invalid.yar contains invalid rule(s)!" in caplog.text
    assert len(results) == 2


@pytest.mark.skipif(not HAS_YARA, reason="requires python-yara")
def test_yara_plugin_compiled_rule(target_yara: Target, tmp_path: str) -> None:
    with tempfile.NamedTemporaryFile(mode="w", dir=tmp_path, delete=False) as tf:
        rules = yara.compile(rule_file)
        rules.save(tf.name)
        tf.close()

        results = list(target_yara.yara(rules=[tf.name]))

        assert len(results) == 2

        assert results[0].path == "/test_file"
        assert results[0].rule == "test_rule_name"
        assert results[0].tags == ["tag1", "tag2", "tag3"]
        assert results[0].namespace == "default"
        assert results[0].digest.md5 == "6f8db599de986fab7a21625b7916589c"
        assert results[0].digest.sha1 == "661295c9cbf9d6b2f6428414504a8deed3020641"
        assert results[0].digest.sha256 == "d5579c46dfcc7f18207013e65b44e4cb4e2c2298f4ac457ba8f82743f31e930b"



import sys
from typing import Iterator
from unittest.mock import MagicMock

import pytest

from dissect.target.target import Target


@pytest.fixture
def mock_impacket(monkeypatch: pytest.MonkeyPatch) -> Iterator[MagicMock]:
    try:
        del sys.modules["dissect.target.loader"]
    except KeyError:
        pass

    with monkeypatch.context() as m:
        mock_impacket = MagicMock()
        m.setitem(sys.modules, "impacket", mock_impacket)
        m.setitem(sys.modules, "impacket.dcerpc", mock_impacket.dcerpc)
        m.setitem(sys.modules, "impacket.dcerpc.v5", mock_impacket.dcerpc.v5)
        m.setitem(sys.modules, "impacket.dcerpc.v5.rpcrt", mock_impacket.dcerpc.v5.rpcrt)
        m.setitem(sys.modules, "impacket.nt_errors", mock_impacket.nt_errors)
        m.setitem(sys.modules, "impacket.smb", mock_impacket.smb)
        m.setitem(sys.modules, "impacket.smb3structs", mock_impacket.smb3structs)
        m.setitem(sys.modules, "impacket.smbconnection", mock_impacket.smbconnection)

        yield mock_impacket


@pytest.fixture
def mock_connection(mock_impacket: MagicMock) -> MagicMock:
    mock_connection = MagicMock()
    mock_impacket.smbconnection.SMBConnection.return_value = mock_connection

    yield mock_connection


def test_smb_loader(mock_impacket: MagicMock, mock_connection: MagicMock) -> None:
    from dissect.target.filesystems.smb import SmbFilesystem
    from dissect.target.loader import open as loader_open
    from dissect.target.loaders.smb import SmbLoader, SmbRegistry

    loader = loader_open("smb://user@host")
    assert isinstance(loader, SmbLoader)
    assert loader._conn is mock_connection
    mock_impacket.smbconnection.SMBConnection.assert_called_with(
        remoteName="host", remoteHost="host", myName=SmbLoader.MACHINE_NAME
    )
    mock_connection.login.assert_called_with(
        domain=".", user="user", password="", nthash=SmbLoader.EMPTY_NT, lmhash=SmbLoader.EMPTY_LM
    )

    mock_connection.listShares.return_value = [{"shi1_netname": "C$\x00"}]

    t = Target()
    loader.map(t)

    assert len(t.filesystems) == 1
    assert isinstance(t.fs.mounts["c:"], SmbFilesystem)
    assert t.fs.mounts["c:"].share_name == "C$"

    assert len(t._plugins) == 1
    assert isinstance(t._plugins[0], SmbRegistry)



from __future__ import annotations

import logging
import re
import uuid
from pathlib import Path
from typing import Iterator

from flow.record.fieldtypes import posix_path

from dissect.target.filesystem import Filesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import UnixUserRecord
from dissect.target.helpers.utils import parse_options_string
from dissect.target.plugin import OperatingSystem, OSPlugin, arg, export
from dissect.target.target import Target

log = logging.getLogger(__name__)


# https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#ISA
ARCH_MAP = {
    0x00: "unknown",
    0x02: "sparc",
    0x03: "x86",
    0x08: "mips",
    0x14: "powerpc32",
    0x15: "powerpc64",
    0x16: "s390",  # and s390x
    0x28: "aarch32",  # armv7
    0x2A: "superh",
    0x32: "ia-64",
    0x3E: "x86_64",
    0xB7: "aarch64",  # armv8
    0xF3: "riscv64",
    0xF7: "bpf",
}


class UnixPlugin(OSPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self._add_mounts()
        self._add_devices()
        self._hostname_dict = self._parse_hostname_string()
        self._hosts_dict = self._parse_hosts_string()
        self._os_release = self._parse_os_release()

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if fs.exists("/var") and fs.exists("/etc"):
                return fs

        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> UnixPlugin:
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(record=UnixUserRecord)
    @arg("--sessions", action="store_true", help="Parse syslog for recent user sessions")
    def users(self, sessions: bool = False) -> Iterator[UnixUserRecord]:
        """Yield unix user records from passwd files or syslog session logins.

        Resources:
            - https://manpages.ubuntu.com/manpages/oracular/en/man5/passwd.5.html
        """

        PASSWD_FILES = ["/etc/passwd", "/etc/passwd-", "/etc/master.passwd"]

        seen_users = set()

        # Yield users found in passwd files.
        for passwd_file in PASSWD_FILES:
            if (path := self.target.fs.path(passwd_file)).exists():
                for line in path.open("rt"):
                    line = line.strip()
                    if not line or line.startswith("#"):
                        continue

                    pwent = dict(enumerate(line.split(":")))

                    current_user = (pwent.get(0), pwent.get(5), pwent.get(6))
                    if current_user in seen_users:
                        continue

                    seen_users.add(current_user)
                    yield UnixUserRecord(
                        name=pwent.get(0),
                        passwd=pwent.get(1),
                        uid=pwent.get(2),
                        gid=pwent.get(3),
                        gecos=pwent.get(4),
                        home=posix_path(pwent.get(5)),
                        shell=pwent.get(6),
                        source=passwd_file,
                        _target=self.target,
                    )

        # Find users not in passwd files by parsing recent
        # syslog ldap, kerberos and x-session logins.
        # Must be enabled using the --sessions flag
        if sessions and (path := self.target.fs.path("/var/log/syslog")).exists():
            sessions = []
            cur_session = -1
            needles = {
                ("setting HOME=", "home"),
                ("setting SHELL=", "shell"),
                ("setting USER=", "name"),
            }

            for line in path.open("rt"):
                # Detect the beginning of a new session activation in the syslog
                #
                # dbus-update-activation-environment starts a new session with
                # DBUS_SESSION_BUS_ADDRESS, XDG_RUNTIME_DIR and/or DISPLAY
                # Using DBUS_SESSION_BUS_ADDRESS seems to work fine for now.
                if "setting DBUS_SESSION_BUS_ADDRESS=" in line:
                    bus = line.split("DBUS_SESSION_BUS_ADDRESS=")[1].strip()
                    cur_session += 1
                    sessions.append({"bus": bus, "name": None, "home": None, "shell": None, "uid": None})

                # Search the following lines for more information on the previousley detected session.
                for n, k in needles:
                    if n in line and "dbus-update-activation-environment" in line:
                        sessions[cur_session][k] = line.split(n)[1].strip()

            for user in sessions:
                # Only return users we didn't already find in previously parsed passwd files and past sessions.
                current_user = (user["name"], user["home"], user["shell"])
                if current_user in seen_users:
                    continue

                seen_users.add(current_user)

                yield UnixUserRecord(
                    name=user["name"],
                    home=posix_path(user["home"]),
                    shell=user["shell"],
                    source="/var/log/syslog",
                    _target=self.target,
                )

    @export(property=True)
    def architecture(self) -> str | None:
        return self._get_architecture(self.os)

    @export(property=True)
    def hostname(self) -> str | None:
        hosts_string = self._hosts_dict.get("hostname", "localhost")
        return self._hostname_dict.get("hostname", hosts_string)

    @export(property=True)
    def domain(self) -> str | None:
        domain = self._hostname_dict.get("domain", "localhost")
        if domain == "localhost":
            domain = self._hosts_dict["hostname", "localhost"]
            if domain == self.hostname:
                return domain  # domain likely not defined, so localhost is the domain.
        return domain

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.UNIX.value

    def _parse_rh_legacy(self, path: Path) -> str | None:
        hostname = None
        file_contents = path.open("rt").readlines()
        for line in file_contents:
            if not line.startswith("HOSTNAME"):
                continue
            _, _, hostname = line.rstrip().partition("=")
        return hostname

    def _parse_hostname_string(self, paths: list[str] | None = None) -> dict[str, str] | None:
        """
        Returns a dict containing the hostname and domain name portion of the path(s) specified

        Args:
            paths (list): list of paths
        """
        redhat_legacy_path = "/etc/sysconfig/network"
        paths = paths or ["/etc/hostname", "/etc/HOSTNAME", redhat_legacy_path]
        hostname_dict = {"hostname": None, "domain": None}

        for path in paths:
            path = self.target.fs.path(path)

            if not path.exists():
                continue

            if path.as_posix() == redhat_legacy_path:
                hostname_string = self._parse_rh_legacy(path)
            else:
                hostname_string = path.open("rt").read().rstrip()

            if hostname_string and "." in hostname_string:
                hostname_string = hostname_string.split(".", maxsplit=1)
                hostname_dict = {"hostname": hostname_string[0], "domain": hostname_string[1]}
            elif hostname_string != "":
                hostname_dict = {"hostname": hostname_string, "domain": None}
            else:
                hostname_dict = {"hostname": None, "domain": None}
            break  # break whenever a valid hostname is found
        return hostname_dict

    def _parse_hosts_string(self, paths: list[str] | None = None) -> dict[str, str]:
        paths = paths or ["/etc/hosts"]
        hosts_string = {"ip": None, "hostname": None}

        for path in paths:
            for fs in self.target.filesystems:
                if fs.exists(path):
                    for line in fs.path(path).open("rt").readlines():
                        line = line.split()
                        if not line:
                            continue

                        if (line[0].startswith("127.0.") or line[0].startswith("::1")) and line[
                            1
                        ].lower() != "localhost":
                            hosts_string = {"ip": line[0], "hostname": line[1]}
        return hosts_string

    def _add_mounts(self) -> None:
        fstab = self.target.fs.path("/etc/fstab")

        for dev_id, volume_name, mount_point, _, options in parse_fstab(fstab, self.target.log):
            opts = parse_options_string(options)
            subvol = opts.get("subvol", None)
            subvolid = opts.get("subvolid", None)
            for fs in self.target.filesystems:
                fs_id = None
                fs_subvol = None
                fs_subvolid = None
                fs_last_mount = None
                fs_volume_name = None
                vol_volume_name = fs.volume.name if fs.volume and not isinstance(fs.volume, list) else None

                if fs.__type__ == "xfs":
                    fs_id = fs.xfs.uuid
                    fs_volume_name = fs.xfs.name
                elif fs.__type__ == "ext":
                    fs_id = fs.extfs.uuid
                    fs_last_mount = fs.extfs.last_mount
                    fs_volume_name = fs.extfs.volume_name
                elif fs.__type__ == "btrfs":
                    fs_id = fs.btrfs.uuid
                    fs_subvol = fs.subvolume.path
                    fs_subvolid = fs.subvolume.objectid
                elif fs.__type__ == "fat":
                    fs_id = fs.fatfs.volume_id
                    # This normalizes fs_id to comply with libblkid generated UUIDs
                    # This is needed because FAT filesystems don't have a real UUID,
                    # but instead a volume_id which is not case-sensitive
                    fs_id = fs_id[:4].upper() + "-" + fs_id[4:].upper()

                if (
                    (fs_id and (fs_id == dev_id and (subvol == fs_subvol or subvolid == fs_subvolid)))
                    or (fs_last_mount and (fs_last_mount == mount_point))
                    or (fs_volume_name and (fs_volume_name == volume_name))
                    or (vol_volume_name and (vol_volume_name == volume_name))
                ):
                    self.target.log.debug("Mounting %s (%s) at %s", fs, fs.volume, mount_point)
                    self.target.fs.mount(mount_point, fs)

    def _add_devices(self) -> None:
        """Add some virtual block devices to the target.

        Currently only adds LVM devices.
        """
        vfs = self.target.fs.append_layer()

        for volume in self.target.volumes:
            if volume.vs and volume.vs.__type__ == "lvm":
                vfs.map_file_fh(f"/dev/{volume.raw.vg.name}/{volume.raw.name}", volume)

    def _parse_os_release(self, glob: str | None = None) -> dict[str, str]:
        """Parse files containing Unix version information.

        Not all these files are equal. Generally speaking these files are
        either key=value files or contain just one line.

        Examples of key=value pair structured release files are:
        - /etc/os-release
        - /usr/lib/os-release
        - /etc/lsb-release

        Examples of sparse release files:
        - /etc/fedora-release
        - /etc/centos-release
        - /etc/redhat-release
        - /etc/SuSE-release

        Examples of bsd version files:
        - /bin/freebsd-version
        """
        glob = glob or "/etc/*-release"

        os_release = {}

        for path in self.target.fs.glob(glob):
            if self.target.fs.path(path).is_file():
                with self.target.fs.path(path).open("rt") as release_file:
                    for line in release_file:
                        if line.startswith("#"):
                            continue

                        elif "=" not in line:
                            os_release["DISTRIB_DESCRIPTION"] = line.strip()

                        else:
                            try:
                                name, value = line.split("=", maxsplit=1)
                                os_release[name] = value.replace('"', "").replace("\n", "")
                            except ValueError:
                                continue
        return os_release

    def _get_architecture(self, os: str = "unix", path: Path | str = "/bin/ls") -> str | None:
        """Determine architecture by reading an ELF header of a binary on the target.

        Resources:
            - https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#ISA
        """

        if not isinstance(path, TargetPath):
            for fs in [self.target.fs, *self.target.filesystems]:
                if (path := fs.path(path)).exists():
                    break

        if not path.exists():
            return

        fh = path.open("rb")
        fh.seek(4)  # ELF - e_ident[EI_CLASS]
        bits = fh.read(1)[0]

        fh.seek(18)  # ELF - e_machine
        e_machine = int.from_bytes(fh.read(2), "little")
        arch = ARCH_MAP.get(e_machine, "unknown")

        return f"{arch}_32-{os}" if bits == 1 and not arch[-2:] == "32" else f"{arch}-{os}"


def parse_fstab(
    fstab: TargetPath,
    log: logging.Logger = log,
) -> Iterator[tuple[uuid.UUID | str, str, str, str, str]]:
    """Parse fstab file and return a generator that streams the details of entries,
    with unsupported FS types and block devices filtered away.
    """

    SKIP_FS_TYPES = (
        "swap",
        "tmpfs",
        "devpts",
        "sysfs",
        "procfs",
        "overlayfs",
    )

    if not fstab.exists():
        return

    for entry in fstab.open("rt"):
        entry = entry.strip()
        if entry.startswith("#"):
            continue

        entry_parts = re.split(r"\s+", entry)

        if len(entry_parts) != 6:
            continue

        dev, mount_point, fs_type, options, _, _ = entry_parts

        if fs_type in SKIP_FS_TYPES:
            log.warning("Skipped FS type: %s, %s, %s", fs_type, dev, mount_point)
            continue

        dev_id = None
        volume_name = None
        if dev.startswith(("/dev/mapper", "/dev/gpt")):
            volume_name = dev.rsplit("/")[-1]
        elif dev.startswith("/dev/disk/by-uuid"):
            dev_id = dev.rsplit("/")[-1]
        elif dev.startswith("/dev/") and dev.count("/") == 3:
            # When composing a vg-lv name, LVM2 replaces hyphens with double hyphens in the vg and lv names
            # Emulate that here when combining the vg and lv names
            volume_name = "-".join(part.replace("-", "--") for part in dev.rsplit("/")[-2:])
        elif dev.startswith("UUID="):
            dev_id = dev.split("=")[1]
        elif dev.startswith("LABEL="):
            volume_name = dev.split("=")[1]
        else:
            log.warning("Unsupported mount device: %s %s", dev, mount_point)
            continue

        if mount_point == "/":
            continue

        if dev_id:
            try:
                dev_id = uuid.UUID(dev_id)
            except ValueError:
                pass

        yield dev_id, volume_name, mount_point, fs_type, options



import json
import os
import re
from typing import Any, Optional
from unittest.mock import MagicMock, patch

import pytest

from dissect.target.plugin import PluginFunction
from dissect.target.target import Target
from dissect.target.tools.query import main as target_query


def test_target_query_list(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-query", "--list"])

        with pytest.raises((SystemExit, IndexError, ImportError)):
            target_query()
        out, _ = capsys.readouterr()

        assert out.startswith("Available plugins:")
        assert "Failed to load:\n    None\nAvailable loaders:\n" in out


@pytest.mark.parametrize(
    "given_funcs, expected_invalid_funcs",
    [
        (
            ["foo"],
            ["foo"],
        ),
        (
            ["bar", "version"],
            ["bar"],
        ),
        (
            ["version", "foo", "wireguard.config", "bar", "apps.webserver.iis*"],
            ["bar", "foo"],
        ),
        (
            ["apps.browser.*.downloads", "bar", "version", "foo"],
            ["bar", "foo"],
        ),
        (
            ["apps.webserver.iis.doesnt.exist", "apps.webserver.apache.access"],
            ["apps.webserver.iis.doesnt.exist*"],
        ),
    ],
)
def test_target_query_invalid_functions(
    capsys: pytest.CaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    given_funcs: list[str],
    expected_invalid_funcs: list[str],
) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-query", "-f", ",".join(given_funcs), "tests/_data/loaders/tar/test-archive.tar.gz"],
        )

        with pytest.raises((SystemExit)):
            target_query()
        _, err = capsys.readouterr()

        assert "target-query: error: argument -f/--function contains invalid plugin(s):" in err

        # Workaround for https://github.com/fox-it/dissect.target/issues/266
        RE_ERR = re.compile(r"contains invalid plugin\(s\)\: (?P<funcs>.*?)\n$")
        match = RE_ERR.search(err).groupdict()

        invalid_funcs = [i.strip() for i in match["funcs"].split(",")]
        invalid_funcs.sort()

        assert invalid_funcs == expected_invalid_funcs


@pytest.mark.parametrize(
    "given_funcs, expected_invalid_funcs",
    [
        (
            ["foo"],
            ["foo"],
        ),
        (
            ["bar", "version"],
            ["bar"],
        ),
        (
            ["version", "foo", "wireguard.config", "bar", "apps.webserver.iis*"],
            ["bar", "foo"],
        ),
        (
            ["apps.browser.*.downloads", "bar", "version", "foo"],
            ["bar", "foo"],
        ),
        (
            ["apps.webserver.iis.doesnt.exist", "apps.webserver.apache.access"],
            ["apps.webserver.iis.doesnt.exist*"],
        ),
    ],
)
def test_target_query_invalid_excluded_functions(
    capsys: pytest.CaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    given_funcs: list[str],
    expected_invalid_funcs: list[str],
) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-query",
                "-f",
                "hostname",
                "-xf",
                ",".join(given_funcs),
                "tests/_data/loaders/tar/test-archive.tar.gz",
            ],
        )

        with pytest.raises((SystemExit)):
            target_query()
        _, err = capsys.readouterr()

        assert "target-query: error: argument -xf/--excluded-functions contains invalid plugin(s):" in err

        # Workaround for https://github.com/fox-it/dissect.target/issues/266
        RE_ERR = re.compile(r"contains invalid plugin\(s\)\: (?P<funcs>.*?)\n$")
        match = RE_ERR.search(err).groupdict()

        invalid_funcs = [i.strip() for i in match["funcs"].split(",")]
        invalid_funcs.sort()

        assert invalid_funcs == expected_invalid_funcs


def test_target_query_unsupported_plugin_log(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-query", "-f", "regf", "tests/_data/loaders/tar/test-archive.tar.gz"],
        )

        target_query()
        _, err = capsys.readouterr()

        assert "Unsupported plugin for regf: Registry plugin not loaded" in err


def mock_find_plugin_function(
    target: Target,
    patterns: str,
    compatibility: bool = False,
    **kwargs,
) -> tuple[list[PluginFunction], set[str]]:
    plugins = []
    for pattern in patterns.split(","):
        plugins.append(
            PluginFunction(
                name=pattern,
                output_type="record",
                path=pattern,
                class_object=MagicMock(),
                method_name=pattern,
                plugin_desc={},
            ),
        )

    return (plugins, set())


def mock_execute_function(
    target: Target,
    func: PluginFunction,
    cli_params: Optional[list[str]] = None,
) -> tuple[str, Any, list[str]]:
    return (func.output_type, func.name, "")


def test_target_query_filtered_functions(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-query",
                "-f",
                "foo,bar,bla,foo",
                "-xf",
                "bla",
                "tests/_data/loaders/tar/test-archive.tar.gz",
            ],
        )

        with (
            patch(
                "dissect.target.tools.query.find_plugin_functions",
                autospec=True,
                side_effect=mock_find_plugin_function,
            ),
            patch(
                "dissect.target.tools.utils.find_plugin_functions",
                autospec=True,
                side_effect=mock_find_plugin_function,
            ),
            patch(
                "dissect.target.tools.query.execute_function_on_target",
                autospec=True,
                side_effect=mock_execute_function,
            ) as mock_execute,
            patch(
                "dissect.target.tools.query.record_output",
                autospec=True,
            ),
        ):
            target_query()

            assert len(mock_execute.mock_calls) == 2

            executed_func_names = set()
            for call in mock_execute.mock_calls:
                executed_func_names.add(call.args[1].name)
            assert executed_func_names == {"foo", "bar"}


def test_target_query_dry_run(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    if os.sep == "\\":
        target_file = "tests\\_data\\loaders\\tar\\test-archive.tar.gz"
    else:
        target_file = "tests/_data/loaders/tar/test-archive.tar.gz"

    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-query", "-f", "general.*", "--dry-run", target_file],
        )

        target_query()
        out, _ = capsys.readouterr()

        assert out == (
            f"Dry run on: <Target {target_file}>\n"
            "  execute: users (general.default.users)\n"
            "  execute: network.interfaces (general.network.interfaces)\n"
            "  execute: osinfo (general.osinfo.osinfo)\n"
        )


def test_target_query_list_json(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """test if target-query --list --json output is formatted as we expect it to be."""

    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-query", "-l", "-j"])
        with pytest.raises((SystemExit, IndexError, ImportError)):
            target_query()
        out, _ = capsys.readouterr()

    try:
        output = json.loads(out)
    except json.JSONDecodeError:
        pass

    # test the generic structure of the returned dictionary.
    assert isinstance(output, dict), "Could not load JSON output of 'target-query --list --json'"
    assert output["plugins"], "Expected a dictionary of plugins"
    assert output["loaders"], "Expected a dictionary of loaders"
    assert len(output["plugins"]["loaded"]) > 200, "Expected more loaded plugins"
    assert not output["plugins"].get("failed"), "Some plugin(s) failed to initialize"

    def get_plugin(plugins: list[dict], needle: str) -> dict:
        match = [p for p in output["plugins"]["loaded"] if p["name"] == needle]
        return match[0] if match else False

    # general plugin
    users_plugin = get_plugin(output, "users")
    assert users_plugin == {
        "name": "users",
        "description": "Return the users available in the target.",
        "output": "record",
        "path": "general.default.users",
    }

    # namespaced plugin
    plocate_plugin = get_plugin(output, "plocate.locate")
    assert plocate_plugin == {
        "name": "plocate.locate",
        "description": "Yield file and directory names from the plocate.db.",
        "output": "record",
        "path": "os.unix.locate.plocate.locate",
    }

    # regular plugin
    sam_plugin = get_plugin(output, "sam")
    assert sam_plugin == {
        "name": "sam",
        "description": "Dump SAM entries",
        "output": "record",
        "path": "os.windows.credential.sam.sam",
    }
